# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test class :class:`iris._concatenate._CubeSignature`."""

# import iris tests first so that some things can be initialised
# before importing anything else.
import iris.tests as tests

from cf_units import Unit
import numpy as np

from iris._concatenate import _CubeSignature as CubeSignature
from iris.coords import DimCoord
from iris.cube import Cube
from iris.util import new_axis


class Test__coordinate_dim_metadata_equality(tests.IrisTest):
    def setUp(self):
        nt = 10
        data = np.arange(nt, dtype=np.float32)
        cube = Cube(data, standard_name="air_temperature", units="K")
        # Temporal coordinate.
        t_units = Unit("hours since 1970-01-01 00:00:00", calendar="gregorian")
        t_coord = DimCoord(
            points=np.arange(nt), standard_name="time", units=t_units
        )
        cube.add_dim_coord(t_coord, 0)

        # Increasing 1D time-series cube.
        self.series_inc_cube = cube
        self.series_inc = CubeSignature(self.series_inc_cube)

        # Decreasing 1D time-series cube.
        self.series_dec_cube = self.series_inc_cube.copy()
        self.series_dec_cube.remove_coord("time")
        t_tmp = DimCoord(
            points=t_coord.points[::-1], standard_name="time", units=t_units
        )
        self.series_dec_cube.add_dim_coord(t_tmp, 0)
        self.series_dec = CubeSignature(self.series_dec_cube)

        # Scalar 0D time-series cube with scalar time coordinate.
        cube = Cube(0, standard_name="air_temperature", units="K")
        cube.add_aux_coord(
            DimCoord(points=nt, standard_name="time", units=t_units)
        )
        self.scalar_cube = cube

    def test_scalar_non_common_axis(self):
        scalar = CubeSignature(self.scalar_cube)
        self.assertNotEqual(self.series_inc.dim_metadata, scalar.dim_metadata)
        self.assertNotEqual(self.series_dec.dim_metadata, scalar.dim_metadata)

    def test_1d_single_value_common_axis(self):
        # Manually promote scalar time cube to be a 1d cube.
        single = CubeSignature(new_axis(self.scalar_cube, "time"))
        self.assertEqual(self.series_inc.dim_metadata, single.dim_metadata)
        self.assertEqual(self.series_dec.dim_metadata, single.dim_metadata)

    def test_increasing_common_axis(self):
        series_inc = self.series_inc
        series_dec = self.series_dec
        self.assertEqual(series_inc.dim_metadata, series_inc.dim_metadata)
        self.assertNotEqual(series_inc.dim_metadata, series_dec.dim_metadata)

    def test_decreasing_common_axis(self):
        series_inc = self.series_inc
        series_dec = self.series_dec
        self.assertNotEqual(series_dec.dim_metadata, series_inc.dim_metadata)
        self.assertEqual(series_dec.dim_metadata, series_dec.dim_metadata)

    def test_circular(self):
        series_inc = self.series_inc
        circular_cube = self.series_inc_cube.copy()
        circular_cube.coord("time").circular = True
        circular = CubeSignature(circular_cube)
        self.assertNotEqual(circular.dim_metadata, series_inc.dim_metadata)
        self.assertEqual(circular.dim_metadata, circular.dim_metadata)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :meth:`iris.analysis.trajectory.interpolate`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import numpy as np

from iris.coords import AuxCoord, DimCoord
import iris.tests.stock

from iris.analysis.trajectory import interpolate


class TestFailCases(tests.IrisTest):
    @tests.skip_data
    def test_derived_coord(self):
        cube = iris.tests.stock.realistic_4d()
        sample_pts = [("altitude", [0, 10, 50])]
        msg = "'altitude'.*derived coordinates are not allowed"
        with self.assertRaisesRegex(ValueError, msg):
            interpolate(cube, sample_pts, "nearest")

        # Try to request unknown interpolation method.

    def test_unknown_method(self):
        cube = iris.tests.stock.simple_2d()
        sample_point = [("x", 2.8)]
        msg = "Unhandled interpolation.*linekar"
        with self.assertRaisesRegex(ValueError, msg):
            interpolate(cube, sample_point, method="linekar")


class TestNearest(tests.IrisTest):
    # Test interpolation with 'nearest' method.
    # This is basically a wrapper to the routine:
    #   'analysis._interpolate_private._nearest_neighbour_indices_ndcoords'.
    # That has its own test, so we don't test the basic calculation
    # exhaustively here.  Instead we check the way it handles the source and
    # result cubes (especially coordinates).

    def setUp(self):
        cube = iris.tests.stock.simple_3d()
        # Actually, this cube *isn't* terribly realistic, as the lat+lon coords
        # have integer type, which in this case produces some peculiar results.
        # Let's fix that (and not bother to test the peculiar behaviour).
        for coord_name in ("longitude", "latitude"):
            coord = cube.coord(coord_name)
            coord.points = coord.points.astype(float)
        self.test_cube = cube
        # Define coordinates for a single-point testcase.
        y_val, x_val = 0, -90
        # Work out cube indices of the testpoint.
        self.single_point_iy = np.where(
            cube.coord("latitude").points == y_val
        )[0][0]
        self.single_point_ix = np.where(
            cube.coord("longitude").points == x_val
        )[0][0]
        # Use slightly-different values to test nearest-neighbour operation.
        self.single_sample_point = [
            ("latitude", [y_val + 19.23]),
            ("longitude", [x_val - 17.54]),
        ]

    def test_single_point_same_cube(self):
        # Check exact result matching for a single point.
        cube = self.test_cube
        result = interpolate(cube, self.single_sample_point, method="nearest")
        # Check that the result is a single trajectory point, exactly equal to
        # the expected part of the original data.
        self.assertEqual(result.shape[-1], 1)
        result = result[..., 0]
        expected = cube[:, self.single_point_iy, self.single_point_ix]
        self.assertEqual(result, expected)

    def test_multi_point_same_cube(self):
        # Check an exact result for multiple points.
        cube = self.test_cube
        # Use latitude selection to recreate a whole row of the original cube.
        sample_points = [
            ("longitude", [-180, -90, 0, 90]),
            ("latitude", [0, 0, 0, 0]),
        ]
        result = interpolate(cube, sample_points, method="nearest")

        # The result should be identical to a single latitude section of the
        # original, but with modified coords (latitude has 4 repeated zeros).
        expected = cube[:, 1, :]
        # Result 'longitude' is now an aux coord.
        co_x = expected.coord("longitude")
        expected.remove_coord(co_x)
        expected.add_aux_coord(co_x, 1)
        # Result 'latitude' is now an aux coord containing 4*[0].
        expected.remove_coord("latitude")
        co_y = AuxCoord(
            [0, 0, 0, 0], standard_name="latitude", units="degrees"
        )
        expected.add_aux_coord(co_y, 1)
        self.assertEqual(result, expected)

    def test_aux_coord_noninterpolation_dim(self):
        # Check exact result with an aux-coord mapped to an uninterpolated dim.
        cube = self.test_cube
        cube.add_aux_coord(DimCoord([17, 19], long_name="aux0"), 0)

        # The result cube should exactly equal a single source point.
        result = interpolate(cube, self.single_sample_point, method="nearest")
        self.assertEqual(result.shape[-1], 1)
        result = result[..., 0]
        expected = cube[:, self.single_point_iy, self.single_point_ix]
        self.assertEqual(result, expected)

    def test_aux_coord_one_interp_dim(self):
        # Check exact result with an aux-coord over one interpolation dims.
        cube = self.test_cube
        cube.add_aux_coord(AuxCoord([11, 12, 13, 14], long_name="aux_x"), 2)

        # The result cube should exactly equal a single source point.
        result = interpolate(cube, self.single_sample_point, method="nearest")
        self.assertEqual(result.shape[-1], 1)
        result = result[..., 0]
        expected = cube[:, self.single_point_iy, self.single_point_ix]
        self.assertEqual(result, expected)

    def test_aux_coord_both_interp_dims(self):
        # Check exact result with an aux-coord over both interpolation dims.
        cube = self.test_cube
        cube.add_aux_coord(
            AuxCoord(
                [[11, 12, 13, 14], [21, 22, 23, 24], [31, 32, 33, 34]],
                long_name="aux_xy",
            ),
            (1, 2),
        )

        # The result cube should exactly equal a single source point.
        result = interpolate(cube, self.single_sample_point, method="nearest")
        self.assertEqual(result.shape[-1], 1)
        result = result[..., 0]
        expected = cube[:, self.single_point_iy, self.single_point_ix]
        self.assertEqual(result, expected)

    def test_aux_coord_fail_mixed_dims(self):
        # Check behaviour with an aux-coord mapped over both interpolation and
        # non-interpolation dims : not supported.
        cube = self.test_cube
        cube.add_aux_coord(
            AuxCoord(
                [[111, 112, 113, 114], [211, 212, 213, 214]],
                long_name="aux_0x",
            ),
            (0, 2),
        )
        msg = (
            "Coord aux_0x at one x-y position has the shape.*"
            "instead of being a single point"
        )
        with self.assertRaisesRegex(ValueError, msg):
            interpolate(cube, self.single_sample_point, method="nearest")

    def test_metadata(self):
        # Check exact result matching for a single point, with additional
        # attributes and cell-methods.
        cube = self.test_cube
        cube.attributes["ODD_ATTR"] = "string-value-example"
        cube.add_cell_method(iris.coords.CellMethod("mean", "area"))
        result = interpolate(cube, self.single_sample_point, method="nearest")
        # Check that the result is a single trajectory point, exactly equal to
        # the expected part of the original data.
        self.assertEqual(result.shape[-1], 1)
        result = result[..., 0]
        expected = cube[:, self.single_point_iy, self.single_point_ix]
        self.assertEqual(result, expected)


if __name__ == "__main__":
    tests.main()



"""
Ionosphere Space Weather
========================

This space weather example plots a filled contour of rotated pole point
data with a shaded relief image underlay. The plot shows aggregated
vertical electron content in the ionosphere.

The plot exhibits an interesting outline effect due to excluding data
values below a certain threshold.

"""

import matplotlib.pyplot as plt
import numpy.ma as ma

import iris
import iris.plot as iplt
import iris.quickplot as qplt


def main():
    # Load the "total electron content" cube.
    filename = iris.sample_data_path("space_weather.nc")
    cube = iris.load_cube(filename, "total electron content")

    # Explicitly mask negative electron content.
    cube.data = ma.masked_less(cube.data, 0)

    # Plot the cube using one hundred colour levels.
    qplt.contourf(cube, 100)
    plt.title("Total Electron Content")
    plt.xlabel("longitude / degrees")
    plt.ylabel("latitude / degrees")
    plt.gca().stock_img()
    plt.gca().coastlines()

    iplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for :class:`iris.analysis.trajectory.Trajectory`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest import mock

import numpy as np

from iris.analysis.trajectory import Trajectory
from iris.tests.stock import simple_3d, simple_4d_with_hybrid_height


class Test___init__(tests.IrisTest):
    def test_2_points(self):
        # basic 2-seg line along x
        waypoints = [{"lat": 0, "lon": 0}, {"lat": 1, "lon": 2}]
        trajectory = Trajectory(waypoints, sample_count=5)

        self.assertEqual(trajectory.length, np.sqrt(5))
        self.assertEqual(trajectory.sample_count, 5)
        self.assertEqual(
            trajectory.sampled_points,
            [
                {"lat": 0.0, "lon": 0.0},
                {"lat": 0.25, "lon": 0.5},
                {"lat": 0.5, "lon": 1.0},
                {"lat": 0.75, "lon": 1.5},
                {"lat": 1.0, "lon": 2.0},
            ],
        )

    def test_3_points(self):
        # basic 2-seg line along x
        waypoints = [
            {"lat": 0, "lon": 0},
            {"lat": 0, "lon": 1},
            {"lat": 0, "lon": 2},
        ]
        trajectory = Trajectory(waypoints, sample_count=21)

        self.assertEqual(trajectory.length, 2.0)
        self.assertEqual(trajectory.sample_count, 21)
        self.assertEqual(
            trajectory.sampled_points[19],
            {"lat": 0.0, "lon": 1.9000000000000001},
        )

    def test_zigzag(self):
        # 4-seg m-shape
        waypoints = [
            {"lat": 0, "lon": 0},
            {"lat": 1, "lon": 1},
            {"lat": 0, "lon": 2},
            {"lat": 1, "lon": 3},
            {"lat": 0, "lon": 4},
        ]
        trajectory = Trajectory(waypoints, sample_count=33)

        self.assertEqual(trajectory.length, 5.6568542494923806)
        self.assertEqual(trajectory.sample_count, 33)
        self.assertEqual(
            trajectory.sampled_points[31],
            {"lat": 0.12499999999999989, "lon": 3.875},
        )


class Test__get_interp_points(tests.IrisTest):
    def test_basic(self):
        dim_names = "lat"
        waypoints = [{dim_names: 0}, {dim_names: 1}]
        sample_count = 5
        trajectory = Trajectory(waypoints, sample_count=sample_count)
        result = trajectory._get_interp_points()
        expected_points = list(np.linspace(0, 1, sample_count))

        self.assertEqual(len(result), len(waypoints[0]))
        self.assertEqual(len(result[0][1]), sample_count)
        self.assertEqual(result[0][1], expected_points)
        self.assertEqual(result[0][0], dim_names)

    def test_2d(self):
        dim_names = ["lat", "lon"]
        waypoints = [
            {dim_names[0]: 0, dim_names[1]: 0},
            {dim_names[0]: 1, dim_names[1]: 2},
        ]
        sample_count = 5
        trajectory = Trajectory(waypoints, sample_count=sample_count)
        result = trajectory._get_interp_points()

        self.assertEqual(len(result), len(waypoints[0]))
        self.assertEqual(len(result[0][1]), sample_count)
        self.assertEqual(len(result[1][1]), sample_count)
        self.assertIn(result[0][0], dim_names)
        self.assertIn(result[1][0], dim_names)

    def test_3d(self):
        dim_names = ["y", "x", "z"]
        waypoints = [
            {dim_names[0]: 0, dim_names[1]: 0, dim_names[2]: 2},
            {dim_names[0]: 1, dim_names[1]: 2, dim_names[2]: 10},
        ]
        sample_count = 5
        trajectory = Trajectory(waypoints, sample_count=sample_count)
        result = trajectory._get_interp_points()

        self.assertEqual(len(result), len(waypoints[0]))
        self.assertEqual(len(result[0][1]), sample_count)
        self.assertEqual(len(result[1][1]), sample_count)
        self.assertEqual(len(result[2][1]), sample_count)
        self.assertIn(result[0][0], dim_names)
        self.assertIn(result[1][0], dim_names)
        self.assertIn(result[2][0], dim_names)


class Test_interpolate(tests.IrisTest):
    def _result_cube_metadata(self, res_cube):
        dim_names = [c.name() for c in res_cube.dim_coords]
        named_dims = [res_cube.coord_dims(c)[0] for c in res_cube.dim_coords]
        anon_dims = list(set(range(res_cube.ndim)) - set(named_dims))
        anon_dims = None if not len(anon_dims) else anon_dims
        return dim_names, named_dims, anon_dims

    def test_cube__simple_3d(self):
        # Test that an 'index' coord is added to the resultant cube.
        cube = simple_3d()
        waypoints = [
            {"latitude": 40, "longitude": 40},
            {"latitude": 0, "longitude": 0},
        ]
        sample_count = 3
        new_coord_name = "index"
        trajectory = Trajectory(waypoints, sample_count=sample_count)
        result = trajectory.interpolate(cube)

        dim_names, named_dims, anon_dims = self._result_cube_metadata(result)
        new_coord = result.coord(new_coord_name)
        exp_named_dims = [0, 1]

        self.assertEqual(result.ndim, cube.ndim - 1)
        self.assertIn(new_coord_name, dim_names)
        self.assertEqual(named_dims, exp_named_dims)
        self.assertIsNone(anon_dims)
        self.assertEqual(len(new_coord.points), sample_count)

    def test_cube__anon_dim(self):
        cube = simple_4d_with_hybrid_height()
        cube.remove_coord("model_level_number")  # Make cube dim 1 anonymous.
        waypoints = [
            {"grid_latitude": 21, "grid_longitude": 31},
            {"grid_latitude": 23, "grid_longitude": 33},
        ]
        sample_count = 4
        new_coord_name = "index"
        trajectory = Trajectory(waypoints, sample_count=sample_count)
        result = trajectory.interpolate(cube)

        dim_names, named_dims, anon_dims = self._result_cube_metadata(result)
        new_coord = result.coord(new_coord_name)
        exp_named_dims = [0, 2]
        exp_anon_dims = [1]

        self.assertEqual(result.ndim, cube.ndim - 1)
        self.assertIn(new_coord_name, dim_names)
        self.assertEqual(named_dims, exp_named_dims)
        self.assertEqual(anon_dims, exp_anon_dims)
        self.assertEqual(len(new_coord.points), sample_count)

    def test_call(self):
        # Test that :func:`iris.analysis.trajectory.interpolate` is called by
        # `Trajectory.interpolate`.
        cube = simple_3d()
        to_patch = "iris.analysis.trajectory.interpolate"
        waypoints = [
            {"latitude": 40, "longitude": 40},
            {"latitude": 0, "longitude": 0},
        ]
        sample_count = 3
        trajectory = Trajectory(waypoints, sample_count=sample_count)

        with mock.patch(to_patch, return_value=cube) as mock_interpolate:
            trajectory.interpolate(cube)
        mock_interpolate.assert_called_once()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :class:`iris._data_manager.DataManager`.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import copy
from unittest import mock

import numpy as np
import numpy.ma as ma

from iris._data_manager import DataManager
from iris._lazy_data import as_lazy_data


class Test___copy__(tests.IrisTest):
    def test(self):
        dm = DataManager(np.array(0))
        emsg = "Shallow-copy of {!r} is not permitted."
        name = type(dm).__name__
        with self.assertRaisesRegex(copy.Error, emsg.format(name)):
            copy.copy(dm)


class Test___deepcopy__(tests.IrisTest):
    def test(self):
        dm = DataManager(np.array(0))
        method = "iris._data_manager.DataManager._deepcopy"
        return_value = mock.sentinel.return_value
        with mock.patch(method) as mocker:
            mocker.return_value = return_value
            result = copy.deepcopy(dm)
            self.assertEqual(mocker.call_count, 1)
            [args], kwargs = mocker.call_args
            self.assertEqual(kwargs, dict())
            self.assertEqual(len(args), 2)
            expected = [return_value, [dm]]
            for item in args.values():
                self.assertIn(item, expected)
        self.assertIs(result, return_value)


class Test___eq__(tests.IrisTest):
    def setUp(self):
        self.shape = (2, 3, 4)
        self.size = np.prod(self.shape)
        self.real_array = np.arange(self.size, dtype=float).reshape(self.shape)

    def test_real_with_real(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(self.real_array.copy())
        self.assertEqual(dm1, dm2)

    def test_real_with_real_failure(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(np.ones(self.shape))
        self.assertFalse(dm1 == dm2)

    def test_real_with_real__dtype_failure(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(self.real_array.astype(int))
        self.assertFalse(dm1 == dm2)

    def test_real_with_lazy_failure(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(as_lazy_data(self.real_array))
        self.assertFalse(dm1 == dm2)
        self.assertFalse(dm2 == dm1)

    def test_lazy_with_lazy(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array))
        self.assertEqual(dm1, dm2)

    def test_lazy_with_lazy_failure(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array) * 10)
        self.assertFalse(dm1 == dm2)

    def test_lazy_with_lazy__dtype_failure(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array).astype(int))
        self.assertFalse(dm1 == dm2)

    def test_non_DataManager_failure(self):
        dm = DataManager(np.array(0))
        self.assertFalse(dm == 0)


class Test___ne__(tests.IrisTest):
    def setUp(self):
        self.shape = (2, 3, 4)
        self.size = np.prod(self.shape)
        self.real_array = np.arange(self.size, dtype=float).reshape(self.shape)

    def test_real_with_real(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(np.ones(self.shape))
        self.assertNotEqual(dm1, dm2)

    def test_real_with_real_failure(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(self.real_array.copy())
        self.assertFalse(dm1 != dm2)

    def test_real_with_real__dtype(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(self.real_array.astype(int))
        self.assertNotEqual(dm1, dm2)

    def test_real_with_lazy(self):
        dm1 = DataManager(self.real_array)
        dm2 = DataManager(as_lazy_data(self.real_array))
        self.assertNotEqual(dm1, dm2)
        self.assertNotEqual(dm2, dm1)

    def test_lazy_with_lazy(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array) * 10)
        self.assertNotEqual(dm1, dm2)

    def test_lazy_with_lazy_failure(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array))
        self.assertFalse(dm1 != dm2)

    def test_lazy_with_lazy__dtype(self):
        dm1 = DataManager(as_lazy_data(self.real_array))
        dm2 = DataManager(as_lazy_data(self.real_array).astype(int))
        self.assertNotEqual(dm1, dm2)

    def test_non_DataManager(self):
        dm = DataManager(np.array(0))
        self.assertNotEqual(dm, 0)


class Test___repr__(tests.IrisTest):
    def setUp(self):
        self.real_array = np.array(123)
        masked_array = ma.array([0, 1], mask=[0, 1])
        self.lazy_array = as_lazy_data(masked_array)
        self.name = DataManager.__name__

    def test_real(self):
        dm = DataManager(self.real_array)
        result = repr(dm)
        expected = "{}({!r})".format(self.name, self.real_array)
        self.assertEqual(result, expected)

    def test_lazy(self):
        dm = DataManager(self.lazy_array)
        result = repr(dm)
        expected = "{}({!r})".format(self.name, self.lazy_array)
        self.assertEqual(result, expected)


class Test__assert_axioms(tests.IrisTest):
    def setUp(self):
        self.real_array = np.array(0)
        self.lazy_array = as_lazy_data(self.real_array)
        self.dm = DataManager(self.real_array)

    def test_array_none(self):
        self.dm._real_array = None
        emsg = "Unexpected data state, got no lazy and no real data"
        with self.assertRaisesRegex(AssertionError, emsg):
            self.dm._assert_axioms()

    def test_array_all(self):
        self.dm._lazy_array = self.lazy_array
        emsg = "Unexpected data state, got lazy and real data"
        with self.assertRaisesRegex(AssertionError, emsg):
            self.dm._assert_axioms()


class Test__deepcopy(tests.IrisTest):
    def setUp(self):
        self.shape = (2, 3, 4)
        self.size = np.prod(self.shape)
        self.real_array = np.arange(self.size, dtype=float).reshape(self.shape)
        self.memo = dict()

    def test_real(self):
        dm = DataManager(self.real_array)
        result = dm._deepcopy(self.memo)
        self.assertEqual(dm, result)

    def test_lazy(self):
        dm = DataManager(as_lazy_data(self.real_array))
        result = dm._deepcopy(self.memo)
        self.assertEqual(dm, result)

    def test_real_with_real(self):
        dm = DataManager(self.real_array)
        data = self.real_array.copy() * 10
        result = dm._deepcopy(self.memo, data=data)
        expected = DataManager(data)
        self.assertEqual(result, expected)
        self.assertIs(result._real_array, data)

    def test_real_with_lazy(self):
        dm = DataManager(self.real_array)
        data = as_lazy_data(self.real_array) * 10
        result = dm._deepcopy(self.memo, data=data)
        expected = DataManager(data)
        self.assertEqual(result, expected)
        self.assertIs(result._lazy_array, data)

    def test_lazy_with_real(self):
        dm = DataManager(as_lazy_data(self.real_array))
        data = self.real_array.copy() * 10
        result = dm._deepcopy(self.memo, data=data)
        expected = DataManager(data)
        self.assertEqual(result, expected)
        self.assertIs(result._real_array, data)

    def test_lazy_with_lazy(self):
        dm = DataManager(as_lazy_data(self.real_array))
        data = as_lazy_data(self.real_array) * 10
        result = dm._deepcopy(self.memo, data=data)
        expected = DataManager(data)
        self.assertEqual(result, expected)
        self.assertIs(result._lazy_array, data)

    def test_real_with_real_failure(self):
        dm = DataManager(self.real_array)
        emsg = "Cannot copy"
        with self.assertRaisesRegex(ValueError, emsg):
            dm._deepcopy(self.memo, data=np.array(0))

    def test_real_with_lazy_failure(self):
        dm = DataManager(self.real_array)
        emsg = "Cannot copy"
        with self.assertRaisesRegex(ValueError, emsg):
            dm._deepcopy(self.memo, data=as_lazy_data(np.array(0)))

    def test_lazy_with_real_failure(self):
        dm = DataManager(as_lazy_data(self.real_array))
        emsg = "Cannot copy"
        with self.assertRaisesRegex(ValueError, emsg):
            dm._deepcopy(self.memo, data=np.array(0))

    def test_lazy_with_lazy_failure(self):
        dm = DataManager(as_lazy_data(self.real_array))
        emsg = "Cannot copy"
        with self.assertRaisesRegex(ValueError, emsg):
            dm._deepcopy(self.memo, data=as_lazy_data(np.array(0)))


class Test_data__getter(tests.IrisTest):
    def setUp(self):
        shape = (2, 3, 4)
        size = np.prod(shape)
        self.real_array = np.arange(size).reshape(shape)
        self.lazy_array = as_lazy_data(self.real_array)
        self.mask_array = ma.masked_array(self.real_array)
        self.mask_array_masked = self.mask_array.copy()
        self.mask_array_masked[0, 0, 0] = ma.masked
        self.dtype = self.mask_array.dtype
        self.fill_value = self.mask_array.fill_value
        self.lazy_mask_array = as_lazy_data(self.mask_array)
        self.lazy_mask_array_masked = as_lazy_data(self.mask_array_masked)

    def test_with_real_array(self):
        dm = DataManager(self.real_array)
        self.assertFalse(dm.has_lazy_data())
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertIs(result, self.real_array)

    def test_with_lazy_array(self):
        dm = DataManager(self.lazy_array)
        self.assertTrue(dm.has_lazy_data())
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(result, self.real_array)

    def test_with_lazy_mask_array__not_masked(self):
        dm = DataManager(self.lazy_mask_array)
        self.assertTrue(dm.has_lazy_data())
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertIsInstance(result, np.core.ndarray)
        self.assertEqual(dm.dtype, self.dtype)
        self.assertEqual(result.fill_value, self.fill_value)
        self.assertArrayEqual(result, self.real_array)

    def test_with_lazy_mask_array__masked(self):
        dm = DataManager(self.lazy_mask_array_masked)
        self.assertTrue(dm.has_lazy_data())
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertIsInstance(result, ma.MaskedArray)
        self.assertEqual(dm.dtype, self.dtype)
        self.assertEqual(result.fill_value, self.fill_value)
        self.assertArrayEqual(result, self.mask_array_masked)

    def test_with_real_masked_constant(self):
        masked_data = ma.masked_array([666], mask=True, dtype=np.dtype("f8"))
        masked_constant = masked_data[0]
        dm = DataManager(masked_constant)
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertIsInstance(result, ma.MaskedArray)
        self.assertNotIsInstance(result, ma.core.MaskedConstant)
        self.assertMaskedArrayEqual(result, masked_data)

    def test_with_lazy_masked_constant(self):
        masked_data = ma.masked_array([666], mask=True)
        masked_constant = masked_data[0]
        lazy_masked_constant = as_lazy_data(masked_constant)
        dm = DataManager(lazy_masked_constant)
        result = dm.data
        self.assertFalse(dm.has_lazy_data())
        self.assertIsInstance(result, ma.MaskedArray)
        self.assertNotIsInstance(result, ma.core.MaskedConstant)
        self.assertMaskedArrayEqual(result, masked_data)


class Test_data__setter(tests.IrisTest):
    def test_zero_ndim_real_with_scalar_int(self):
        value = 456
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = value
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, np.array(value))

    def test_zero_ndim_real_with_scalar_float(self):
        value = 456.0
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = value
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, np.array(value))

    def test_zero_ndim_real_with_zero_ndim_real(self):
        real_array = np.array(456)
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_zero_ndim_real_with_zero_ndim_lazy(self):
        lazy_array = as_lazy_data(np.array(456))
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_zero_ndim_lazy_with_zero_ndim_real(self):
        real_array = np.array(456)
        dm = DataManager(as_lazy_data(np.array(123)))
        self.assertTrue(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_zero_ndim_lazy_with_zero_ndim_lazy(self):
        lazy_array = as_lazy_data(np.array(456))
        dm = DataManager(as_lazy_data(np.array(123)))
        self.assertTrue(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_zero_ndim_real_to_scalar_1d_real_promote(self):
        real_array = np.array([456])
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_zero_ndim_real_to_scalar_1d_lazy_promote(self):
        lazy_array = as_lazy_data(np.array([456]))
        dm = DataManager(np.array(123))
        self.assertFalse(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_zero_ndim_lazy_to_scalar_1d_real_promote(self):
        real_array = np.array([456])
        dm = DataManager(as_lazy_data(np.array(123)))
        self.assertTrue(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_zero_ndim_lazy_to_scalar_1d_lazy_promote(self):
        lazy_array = as_lazy_data(np.array([456]))
        dm = DataManager(as_lazy_data(np.array(123)))
        self.assertTrue(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_scalar_1d_to_zero_ndim_fail(self):
        dm = DataManager(np.array([123]))
        emsg = r"Require data with shape \(1,\), got \(\)."
        with self.assertRaisesRegex(ValueError, emsg):
            dm.data = 456

    def test_nd_real_to_nd_real(self):
        shape = (2, 3, 4)
        size = np.prod(shape)
        real_array = np.arange(size).reshape(shape)
        dm = DataManager(real_array * 10)
        self.assertFalse(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_nd_real_to_nd_lazy(self):
        shape = (2, 3, 4)
        size = np.prod(shape)
        real_array = np.arange(size).reshape(shape)
        lazy_array = as_lazy_data(real_array) * 10
        dm = DataManager(real_array)
        self.assertFalse(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_nd_lazy_to_nd_real(self):
        shape = (2, 3, 4)
        size = np.prod(shape)
        real_array = np.arange(size).reshape(shape)
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array * 10)
        self.assertTrue(dm.has_lazy_data())
        dm.data = real_array
        self.assertFalse(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, real_array)

    def test_nd_lazy_to_nd_lazy(self):
        shape = (2, 3, 4)
        size = np.prod(shape)
        real_array = np.arange(size).reshape(shape)
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array * 10)
        self.assertTrue(dm.has_lazy_data())
        dm.data = lazy_array
        self.assertTrue(dm.has_lazy_data())
        self.assertArrayEqual(dm.data, lazy_array.compute())

    def test_coerce_to_ndarray(self):
        shape = (2, 3)
        size = np.prod(shape)
        real_array = np.arange(size).reshape(shape)
        matrix = np.matrix(real_array)
        dm = DataManager(real_array)
        dm.data = matrix
        self.assertIsInstance(dm._real_array, np.core.ndarray)
        self.assertIsInstance(dm.data, np.core.ndarray)
        self.assertArrayEqual(dm.data, real_array)

    def test_real_masked_constant_to_array(self):
        masked_data = ma.masked_array([666], mask=True, dtype=np.dtype("f8"))
        masked_constant = masked_data[0]
        dm = DataManager(masked_constant)
        self.assertIsInstance(dm._real_array, ma.MaskedArray)
        self.assertNotIsInstance(dm._real_array, ma.core.MaskedConstant)
        self.assertIsInstance(dm.data, ma.MaskedArray)
        self.assertNotIsInstance(dm.data, ma.core.MaskedConstant)
        self.assertMaskedArrayEqual(dm.data, masked_data)


class Test_dtype(tests.IrisTest):
    def setUp(self):
        self.real_array = np.array(0, dtype=np.dtype("int64"))
        self.lazy_array = as_lazy_data(np.array(0, dtype=np.dtype("float64")))

    def test_real_array(self):
        dm = DataManager(self.real_array)
        self.assertEqual(dm.dtype, np.dtype("int64"))

    def test_lazy_array(self):
        dm = DataManager(self.lazy_array)
        self.assertEqual(dm.dtype, np.dtype("float64"))


class Test_ndim(tests.IrisTest):
    def test_ndim_0(self):
        real_array = np.array(0)
        dm = DataManager(real_array)
        self.assertEqual(dm.ndim, 0)
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array)
        self.assertEqual(dm.ndim, 0)

    def test_ndim_nd(self):
        shape = (2, 3, 4)
        real_array = np.arange(24).reshape(shape)
        dm = DataManager(real_array)
        self.assertEqual(dm.ndim, len(shape))
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array)
        self.assertEqual(dm.ndim, len(shape))


class Test_shape(tests.IrisTest):
    def test_shape_scalar(self):
        real_array = np.array(0)
        dm = DataManager(real_array)
        self.assertEqual(dm.shape, ())
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array)
        self.assertEqual(dm.shape, ())

    def test_shape_nd(self):
        shape = (2, 3, 4)
        real_array = np.arange(24).reshape(shape)
        dm = DataManager(real_array)
        self.assertEqual(dm.shape, shape)
        lazy_array = as_lazy_data(real_array)
        dm = DataManager(lazy_array)
        self.assertEqual(dm.shape, shape)


class Test_copy(tests.IrisTest):
    def setUp(self):
        self.method = "iris._data_manager.DataManager._deepcopy"
        self.data = mock.sentinel.data
        self.return_value = mock.sentinel.return_value
        self.memo = {}

    def test(self):
        dm = DataManager(np.array(0))
        kwargs = dict(data=self.data)
        with mock.patch(self.method) as mocker:
            mocker.return_value = self.return_value
            result = dm.copy(data=self.data)
            mocker.assert_called_once_with(self.memo, **kwargs)
        self.assertIs(result, self.return_value)


class Test_core_data(tests.IrisTest):
    def test_real_array(self):
        real_array = np.array(0)
        dm = DataManager(real_array)
        self.assertIs(dm.core_data(), real_array)

    def test_lazy_array(self):
        lazy_array = as_lazy_data(np.array(0))
        dm = DataManager(lazy_array)
        self.assertIs(dm.core_data(), lazy_array)


class Test_has_lazy_data(tests.IrisTest):
    def setUp(self):
        self.real_array = np.array(0)
        self.lazy_array = as_lazy_data(self.real_array)

    def test_with_lazy_array(self):
        dm = DataManager(self.lazy_array)
        self.assertTrue(dm.has_lazy_data())

    def test_with_real_array(self):
        dm = DataManager(self.real_array)
        self.assertFalse(dm.has_lazy_data())


class Test_lazy_data(tests.IrisTest):
    def setUp(self):
        self.real_array = np.array(0)
        self.lazy_array = as_lazy_data(self.real_array)

    def test_with_real_array(self):
        dm = DataManager(self.real_array)
        self.assertFalse(dm.has_lazy_data())
        result = dm.lazy_data()
        self.assertFalse(dm.has_lazy_data())
        self.assertEqual(result, self.lazy_array)
        self.assertFalse(dm.has_lazy_data())

    def test_with_lazy_array(self):
        dm = DataManager(self.lazy_array)
        self.assertTrue(dm.has_lazy_data())
        result = dm.lazy_data()
        self.assertTrue(dm.has_lazy_data())
        self.assertIs(result, dm._lazy_array)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test function :func:`iris.fileformats._pyke_rules.compiled_krb.\
fc_rules_cf_fc.build_mercator_coordinate_system`.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

from unittest import mock

import iris
from iris.coord_systems import Mercator
from iris.fileformats._pyke_rules.compiled_krb.fc_rules_cf_fc import \
    build_mercator_coordinate_system


class TestBuildMercatorCoordinateSystem(tests.IrisTest):
    def test_valid(self):
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=-90,
            semi_major_axis=6377563.396,
            semi_minor_axis=6356256.909)

        cs = build_mercator_coordinate_system(None, cf_grid_var)

        expected = Mercator(
            longitude_of_projection_origin=(
                cf_grid_var.longitude_of_projection_origin),
            ellipsoid=iris.coord_systems.GeogCS(
                cf_grid_var.semi_major_axis,
                cf_grid_var.semi_minor_axis))
        self.assertEqual(cs, expected)

    def test_inverse_flattening(self):
        cf_grid_var = mock.Mock(
            spec=[],
            longitude_of_projection_origin=-90,
            semi_major_axis=6377563.396,
            inverse_flattening=299.3249646)

        cs = build_mercator_coordinate_system(None, cf_grid_var)

        expected = Mercator(
            longitude_of_projection_origin=(
                cf_grid_var.longitude_of_projection_origin),
            ellipsoid=iris.coord_systems.GeogCS(
                cf_grid_var.semi_major_axis,
                inverse_flattening=cf_grid_var.inverse_flattening))
        self.assertEqual(cs, expected)

    def test_longitude_missing(self):
        cf_grid_var = mock.Mock(
            spec=[],
            semi_major_axis=6377563.396,
            inverse_flattening=299.3249646)

        cs = build_mercator_coordinate_system(None, cf_grid_var)

        expected = Mercator(
            ellipsoid=iris.coord_systems.GeogCS(
                cf_grid_var.semi_major_axis,
                inverse_flattening=cf_grid_var.inverse_flattening))
        self.assertEqual(cs, expected)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Code to implement vector rotation by angles, and inferring gridcell angles
from coordinate points and bounds.

"""

import numpy as np

import cartopy.crs as ccrs
import iris


def _3d_xyz_from_latlon(lon, lat):
    """
    Return locations of (lon, lat) in 3D space.

    Args:

    * lon, lat: (float array)
        Arrays of longitudes and latitudes, in degrees.
        Both the same shape.

    Returns:

    * xyz : (array, dtype=float64)
        Cartesian coordinates on a unit sphere.
        Shape is (3, <input-shape>).
        The x / y / z coordinates are in xyz[0 / 1 / 2].

    """
    lon1 = np.deg2rad(lon).astype(np.float64)
    lat1 = np.deg2rad(lat).astype(np.float64)

    x = np.cos(lat1) * np.cos(lon1)
    y = np.cos(lat1) * np.sin(lon1)
    z = np.sin(lat1)

    result = np.concatenate([array[np.newaxis] for array in (x, y, z)])

    return result


def _latlon_from_xyz(xyz):
    """
    Return arrays of lons+lats angles from xyz locations.

    Args:

    * xyz: (array)
        Array of 3-D cartesian coordinates.
        Shape (3, <input_points_dimensions>).
        x / y / z values are in xyz[0 / 1 / 2],

    Returns:

    * lonlat : (array)
        longitude and latitude position angles, in degrees.
        Shape (2, <input_points_dimensions>).
        The longitudes / latitudes are in lonlat[0 / 1].

    """
    lons = np.rad2deg(np.arctan2(xyz[1], xyz[0]))
    radii = np.sqrt(np.sum(xyz * xyz, axis=0))
    lats = np.rad2deg(np.arcsin(xyz[2] / radii))
    return np.array([lons, lats])


def _angle(p, q, r):
    """
    Estimate grid-angles to true-Eastward direction from positions in the same
    grid row, but at increasing column (grid-Eastward) positions.

    {P, Q, R} are locations of consecutive points in the same grid row.
    These could be successive points in a single grid,
        e.g. {T(i-1,j), T(i,j), T(i+1,j)}
    or a mixture of U/V and T gridpoints if row positions are aligned,
        e.g. {v(i,j), f(i,j), v(i+1,j)}.

    Method:

        Calculate dot product of a unit-vector parallel to P-->R, unit-scaled,
        with the unit eastward (true longitude) vector at Q.
        This value is cos(required angle).
        Discriminate between +/- angles by comparing latitudes of P and R.
        Return NaN where any P-->R are zero.

        .. NOTE::

            This method assumes that the vector PR is parallel to the surface
            at the longitude of Q, as it uses the length of PR as the basis for
            the cosine ratio.
            That is only exact when Q is at the same longitude as the midpoint
            of PR, and this typically causes errors which grow with increasing
            gridcell angle.
            However, we retain this method because it reproduces the "standard"
            gridcell-orientation-angle arrays found in files output by the CICE
            model, which presumably uses an equivalent calculation.

    Args:

    * p, q, r : (float array)
        Arrays of angles, in degrees.
        All the same shape.
        Shape is (2, <input_points_dimensions>).
        Longitudes / latitudes are in array[0 / 1].

    Returns:

    * angle : (float array)
        Grid angles relative to true-East, in degrees.
        Positive when grid-East is anticlockwise from true-East.
        Shape is same as <input_points_dimensions>.

    """
    mid_lons = np.deg2rad(q[0])

    pr = _3d_xyz_from_latlon(r[0], r[1]) - _3d_xyz_from_latlon(p[0], p[1])
    pr_norm = np.sqrt(np.sum(pr ** 2, axis=0))
    pr_top = pr[1] * np.cos(mid_lons) - pr[0] * np.sin(mid_lons)

    index = pr_norm == 0
    pr_norm[index] = 1

    cosine = np.maximum(np.minimum(pr_top / pr_norm, 1), -1)
    cosine[index] = 0

    psi = np.arccos(cosine) * np.sign(r[1] - p[1])
    psi[index] = np.nan

    return np.rad2deg(psi)


def gridcell_angles(x, y=None, cell_angle_boundpoints="mid-lhs, mid-rhs"):
    """
    Calculate gridcell orientations for an arbitrary 2-dimensional grid.

    The input grid is defined by two 2-dimensional coordinate arrays with the
    same dimensions (ny, nx), specifying the geolocations of a 2D mesh.

    Input values may be coordinate points (ny, nx) or bounds (ny, nx, 4).
    However, if points, the edges in the X direction are assumed to be
    connected by wraparound.

    Input can be either two arrays, two coordinates, or a single cube
    containing two suitable coordinates identified with the 'x' and 'y' axes.

    Args:

    The inputs (x [,y]) can be any of the following :

    * x (:class:`~iris.cube.Cube`):
        a grid cube with 2D X and Y coordinates, identified by 'axis'.
        The coordinates must be 2-dimensional with the same shape.
        The two dimensions represent grid dimensions in the order Y, then X.

    * x, y (:class:`~iris.coords.Coord`):
        X and Y coordinates, specifying grid locations on the globe.
        The coordinates must be 2-dimensional with the same shape.
        The two dimensions represent grid dimensions in the order Y, then X.
        If there is no coordinate system, they are assumed to be true
        longitudes and latitudes.  Units must convertible to 'degrees'.

    * x, y (2-dimensional arrays of same shape (ny, nx)):
        longitude and latitude cell center locations, in degrees.
        The two dimensions represent grid dimensions in the order Y, then X.

    * x, y (3-dimensional arrays of same shape (ny, nx, 4)):
        longitude and latitude cell bounds, in degrees.
        The first two dimensions are grid dimensions in the order Y, then X.
        The last index maps cell corners anticlockwise from bottom-left.

    Optional Args:

    * cell_angle_boundpoints (string):
        Controls which gridcell bounds locations are used to calculate angles,
        if the inputs are bounds or bounded coordinates.
        Valid values are 'lower-left, lower-right', which takes the angle from
        the lower left to the lower right corner, and 'mid-lhs, mid-rhs' which
        takes an angles between the average of the left-hand and right-hand
        pairs of corners.  The default is 'mid-lhs, mid-rhs'.

    Returns:

        angles : (2-dimensional cube)

            Cube of angles of grid-x vector from true Eastward direction for
            each gridcell, in degrees.
            It also has "true" longitude and latitude coordinates, with no
            coordinate system.
            When the input has coords, then the output ones are identical if
            the inputs are true-latlons, otherwise they are transformed
            true-latlon versions.
            When the input has bounded coords, then the output coords have
            matching bounds and centrepoints (possibly transformed).
            When the input is 2d arrays, or has unbounded coords, then the
            output coords have matching points and no bounds.
            When the input is 3d arrays, then the output coords have matching
            bounds, and the centrepoints are an average of the 4 boundpoints.

    """
    cube = None
    if hasattr(x, "add_aux_coord"):
        # Passed a cube : extract 'x' and ;'y' axis coordinates.
        cube = x  # Save for later checking.
        x, y = cube.coord(axis="x"), cube.coord(axis="y")

    # Now should have either 2 coords or 2 arrays.
    if not hasattr(x, "shape") or not hasattr(y, "shape"):
        msg = (
            "Inputs (x,y) must have array shape property."
            "Got type(x)={} and type(y)={}."
        )
        raise ValueError(msg.format(type(x), type(y)))

    x_coord, y_coord = None, None
    if hasattr(x, "bounds") and hasattr(y, "bounds"):
        # x and y are Coords.
        x_coord, y_coord = x.copy(), y.copy()

        # They must be angles : convert into degrees
        for coord in (x_coord, y_coord):
            if not coord.units.is_convertible("degrees"):
                msg = (
                    "Input X and Y coordinates must have angular "
                    'units. Got units of "{!s}" and "{!s}".'
                )
                raise ValueError(msg.format(x_coord.units, y_coord.units))
            coord.convert_units("degrees")

        if x_coord.ndim != 2 or y_coord.ndim != 2:
            msg = (
                "Coordinate inputs must have 2-dimensional shape. "
                "Got x-shape of {} and y-shape of {}."
            )
            raise ValueError(msg.format(x_coord.shape, y_coord.shape))
        if x_coord.shape != y_coord.shape:
            msg = (
                "Coordinate inputs must have same shape. "
                "Got x-shape of {} and y-shape of {}."
            )
            raise ValueError(msg.format(x_coord.shape, y_coord.shape))
        if cube:
            x_dims, y_dims = (cube.coord_dims(co) for co in (x, y))
            if x_dims != y_dims:
                msg = (
                    "X and Y coordinates must have the same cube "
                    "dimensions.  Got x-dims = {} and y-dims = {}."
                )
                raise ValueError(msg.format(x_dims, y_dims))
        cs = x_coord.coord_system
        if y_coord.coord_system != cs:
            msg = (
                "Coordinate inputs must have same coordinate system. "
                "Got x of {} and y of {}."
            )
            raise ValueError(msg.format(cs, y_coord.coord_system))

        # Base calculation on bounds if we have them, or points as a fallback.
        if x_coord.has_bounds() and y_coord.has_bounds():
            x, y = x_coord.bounds, y_coord.bounds
        else:
            x, y = x_coord.points, y_coord.points

        # Make sure these arrays are ordinary lats+lons, in degrees.
        if cs is not None:
            # Transform points into true lats + lons.
            crs_src = cs.as_cartopy_crs()
            crs_pc = ccrs.PlateCarree()

            def transform_xy_arrays(x, y):
                # Note: flatten, as transform_points is limited to 2D arrays.
                shape = x.shape
                x, y = (arr.flatten() for arr in (x, y))
                pts = crs_pc.transform_points(crs_src, x, y)
                x = pts[..., 0].reshape(shape)
                y = pts[..., 1].reshape(shape)
                return x, y

            # Transform the main reference points into standard lats+lons.
            x, y = transform_xy_arrays(x, y)

            # Likewise replace the original coordinates with transformed ones,
            # because the calculation also needs the centrepoint values.
            xpts, ypts = (coord.points for coord in (x_coord, y_coord))
            xbds, ybds = (coord.bounds for coord in (x_coord, y_coord))
            xpts, ypts = transform_xy_arrays(xpts, ypts)
            xbds, ybds = transform_xy_arrays(xbds, ybds)
            x_coord = iris.coords.AuxCoord(
                points=xpts,
                bounds=xbds,
                standard_name="longitude",
                units="degrees",
            )
            y_coord = iris.coords.AuxCoord(
                points=ypts,
                bounds=ybds,
                standard_name="latitude",
                units="degrees",
            )

    elif hasattr(x, "bounds") or hasattr(y, "bounds"):
        # One was a Coord, and the other not ?
        is_and_not = ("x", "y")
        if hasattr(y, "bounds"):
            is_and_not = reversed(is_and_not)
        msg = "Input {!r} is a Coordinate, but {!r} is not."
        raise ValueError(msg.format(*is_and_not))

    # Now have either 2 points arrays (ny, nx) or 2 bounds arrays (ny, nx, 4).
    # Construct (lhs, mid, rhs) where these represent 3 points at increasing
    # grid-x indices (columns).
    # Also make suitable X and Y coordinates for the result cube.
    if x.ndim == 2:
        # Data is points arrays.
        # Use previous + subsequent points along grid-x-axis as references.

        # PROBLEM: we assume that the rhs connects to the lhs, so we should
        # really only use this if data is full-longitudes (as a 'circular'
        # coordinate).
        # This is mentioned in the docstring, but we have no general means of
        # checking it.

        # NOTE: we take the 2d grid as presented, so the second dimension is
        # the 'X' dim.  Again, that is implicit + can't be checked.
        mid = np.array([x, y])
        lhs = np.roll(mid, 1, 2)
        rhs = np.roll(mid, -1, 2)
        if not x_coord:
            # Create coords for result cube : with no bounds.
            y_coord = iris.coords.AuxCoord(
                x, standard_name="latitude", units="degrees"
            )
            x_coord = iris.coords.AuxCoord(
                y, standard_name="longitude", units="degrees"
            )
    else:
        # Data is bounds arrays.
        # Use gridcell corners at different grid-x positions as references.
        # NOTE: so with bounds, we *don't* need full circular longitudes.
        xyz = _3d_xyz_from_latlon(x, y)
        # Support two different choices of reference points locations.
        angle_boundpoints_vals = {
            "mid-lhs, mid-rhs": "03_to_12",
            "lower-left, lower-right": "0_to_1",
        }
        bounds_pos = angle_boundpoints_vals.get(cell_angle_boundpoints)
        if bounds_pos == "0_to_1":
            lhs_xyz = xyz[..., 0]
            rhs_xyz = xyz[..., 1]
        elif bounds_pos == "03_to_12":
            lhs_xyz = 0.5 * (xyz[..., 0] + xyz[..., 3])
            rhs_xyz = 0.5 * (xyz[..., 1] + xyz[..., 2])
        else:
            msg = (
                'unrecognised cell_angle_boundpoints of "{}", '
                "must be one of {}"
            )
            raise ValueError(
                msg.format(
                    cell_angle_boundpoints, list(angle_boundpoints_vals.keys())
                )
            )
        if not x_coord:
            # Create bounded coords for result cube.
            # Use average of lhs+rhs points in 3d to get 'mid' points,
            # as coords without points are not allowed.
            mid_xyz = 0.5 * (lhs_xyz + rhs_xyz)
            mid_latlons = _latlon_from_xyz(mid_xyz)
            # Create coords with given bounds, and averaged centrepoints.
            x_coord = iris.coords.AuxCoord(
                points=mid_latlons[0],
                bounds=x,
                standard_name="longitude",
                units="degrees",
            )
            y_coord = iris.coords.AuxCoord(
                points=mid_latlons[1],
                bounds=y,
                standard_name="latitude",
                units="degrees",
            )

        # Convert lhs and rhs points back to latlon form -- IN DEGREES !
        lhs = _latlon_from_xyz(lhs_xyz)
        rhs = _latlon_from_xyz(rhs_xyz)
        # 'mid' is coord.points, whether from input or just made up.
        mid = np.array([x_coord.points, y_coord.points])

    # Do the angle calcs, and return as a suitable cube.
    angles = _angle(lhs, mid, rhs)
    result = iris.cube.Cube(
        angles, long_name="gridcell_angle_from_true_east", units="degrees"
    )
    result.add_aux_coord(x_coord, (0, 1))
    result.add_aux_coord(y_coord, (0, 1))
    return result


def rotate_grid_vectors(
    u_cube, v_cube, grid_angles_cube=None, grid_angles_kwargs=None
):
    """
    Rotate distance vectors from grid-oriented to true-latlon-oriented.

    Can also rotate by arbitrary angles, if they are passed in.

    .. Note::

        This operation overlaps somewhat in function with
        :func:`iris.analysis.cartography.rotate_winds`.
        However, that routine only rotates vectors according to transformations
        between coordinate systems.
        This function, by contrast, can rotate vectors by arbitrary angles.
        Most commonly, the angles are estimated solely from grid sampling
        points, using :func:`gridcell_angles` :  This allows operation on
        complex meshes defined by two-dimensional coordinates, such as most
        ocean grids.

    Args:

    * u_cube, v_cube : (cube)
        Cubes of grid-u and grid-v vector components.
        Units should be differentials of true-distance, e.g. 'm/s'.

    Optional args:

    * grid_angles_cube : (cube)
        gridcell orientation angles.
        Units must be angular, i.e. can be converted to 'radians'.
        If not provided, grid angles are estimated from 'u_cube' using the
        :func:`gridcell_angles` method.

    * grid_angles_kwargs : (dict or None)
        Additional keyword args to be passed to the :func:`gridcell_angles`
        method, if it is used.

    Returns:

        true_u, true_v : (cube)
            Cubes of true-north oriented vector components.
            Units are same as inputs.

        .. Note::

            Vector magnitudes will always be the same as the inputs.

    """
    u_out, v_out = (cube.copy() for cube in (u_cube, v_cube))
    if not grid_angles_cube:
        grid_angles_kwargs = grid_angles_kwargs or {}
        grid_angles_cube = gridcell_angles(u_cube, **grid_angles_kwargs)
    gridangles = grid_angles_cube.copy()
    gridangles.convert_units("radians")
    uu, vv, aa = (cube.data for cube in (u_out, v_out, gridangles))
    mags = np.sqrt(uu * uu + vv * vv)
    angs = np.arctan2(vv, uu) + aa
    uu, vv = mags * np.cos(angs), mags * np.sin(angs)

    # Promote all to masked arrays, and also apply mask at bad (NaN) angles.
    mask = np.isnan(aa)
    for cube in (u_out, v_out, aa):
        if hasattr(cube.data, "mask"):
            mask |= cube.data.mask
    u_out.data = np.ma.masked_array(uu, mask=mask)
    v_out.data = np.ma.masked_array(vv, mask=mask)

    return u_out, v_out



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot._get_plot_defn` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import iris.coords
from iris.tests.stock import simple_2d, simple_2d_w_multidim_coords

if tests.MPL_AVAILABLE:
    import iris.plot as iplt


@tests.skip_plot
class Test_get_plot_defn(tests.IrisTest):
    def test_axis_order_xy(self):
        cube_xy = simple_2d()
        defn = iplt._get_plot_defn(cube_xy, iris.coords.POINT_MODE)
        self.assertEqual(
            [coord.name() for coord in defn.coords], ["bar", "foo"]
        )

    def test_axis_order_yx(self):
        cube_yx = simple_2d()
        cube_yx.transpose()
        defn = iplt._get_plot_defn(cube_yx, iris.coords.POINT_MODE)
        self.assertEqual(
            [coord.name() for coord in defn.coords], ["foo", "bar"]
        )

    def test_2d_coords(self):
        cube = simple_2d_w_multidim_coords()
        defn = iplt._get_plot_defn(cube, iris.coords.BOUND_MODE)
        self.assertEqual(
            [coord.name() for coord in defn.coords], ["bar", "foo"]
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Experimental code can be introduced to Iris through this package.

Changes to experimental code may be more extensive than in the rest of the
codebase. The code is expected to graduate, eventually, to "full status".

"""



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Extra stock routines for making and manipulating cubes with 2d coordinates,
to mimic ocean grid data.

"""

import numpy as np
import numpy.ma as ma

from iris.analysis.cartography import unrotate_pole
from iris.cube import Cube
from iris.coords import AuxCoord, DimCoord
from iris.coord_systems import RotatedGeogCS


def expand_1d_x_and_y_bounds_to_2d_xy(x_bounds_1d, y_bounds_1d):
    """
    Convert bounds for separate 1-D X and Y coords into bounds for the
    equivalent 2D coordinates.

    The output arrays have 4 points per cell, for 4 'corners' of a gridcell,
    in the usual anticlockwise order
    (bottom-left, bottom-right, top-right, top-left).

    If 1-dimensional X and Y coords have shapes nx and ny, then their bounds
    have shapes  (nx, 2) and (ny, 2).
    The equivalent 2d coordinates would have values which are a "meshgrid" of
    the original 1-D points, both having the shape (ny, ny).
    The outputs are 2d bounds arrays suitable for such 2d coordinates.

    Args:

    * x_bounds_1d, y_bounds_1d : (array)
        Coordinate bounds arrays, with shapes (nx, 2) and (ny, 2).

    Result:

    * x_bounds_2d, y_bounds_2d : (array)
        Expanded 2d bounds arrays, both of shape (ny, nx, 4).

    """
    shapes = [bds.shape for bds in (x_bounds_1d, y_bounds_1d)]
    for shape in shapes:
        if len(shape) != 2 or shape[1] != 2:
            msg = (
                "One-dimensional bounds arrays must have shapes (ny, 2) "
                "and (nx, 2).  Got {} and {}."
            )
            raise ValueError(msg.format(*shapes))

    # Construct output arrays, which are both (ny, nx, 4).
    nx, ny = [shape[0] for shape in shapes]
    bds_2d_x = np.zeros((ny, nx, 4))
    bds_2d_y = bds_2d_x.copy()

    # Expand x bounds to 2D array (ny, nx, 4) : the same over all 'Y'.
    # Bottom left+right corners are the original 1-D x bounds.
    bds_2d_x[:, :, 0] = x_bounds_1d[:, 0].reshape((1, nx))
    bds_2d_x[:, :, 1] = x_bounds_1d[:, 1].reshape((1, nx))
    # Top left+right corners are the same as bottom left+right.
    bds_2d_x[:, :, 2] = bds_2d_x[:, :, 1].copy()
    bds_2d_x[:, :, 3] = bds_2d_x[:, :, 0].copy()

    # Expand y bounds to 2D array (ny, nx, 4) : the same over all 'X'.
    # Left-hand lower+upper corners are the original 1-D y bounds.
    bds_2d_y[:, :, 0] = y_bounds_1d[:, 0].reshape((ny, 1))
    bds_2d_y[:, :, 3] = y_bounds_1d[:, 1].reshape((ny, 1))
    # Right-hand lower+upper corners are the same as the left-hand ones.
    bds_2d_y[:, :, 1] = bds_2d_y[:, :, 0].copy()
    bds_2d_y[:, :, 2] = bds_2d_y[:, :, 3].copy()

    return bds_2d_x, bds_2d_y


def grid_coords_2d_from_1d(x_coord_1d, y_coord_1d):
    """
    Calculate a pair of 2d X+Y coordinates from 1d ones, in a "meshgrid" style.
    If the inputs are bounded, the outputs have 4 points per bounds in the
    usual way, i.e. points 0,1,2,3 are the gridcell corners anticlockwise from
    the bottom left.

    """
    for coord in (x_coord_1d, y_coord_1d):
        if coord.ndim != 1:
            msg = (
                "Input coords must be one-dimensional. "
                'Coordinate "{}" has shape {}.'
            )
            raise ValueError(msg.format(coord.name(), coord.shape))

    # Calculate centre-points as a mesh of the 2 inputs.
    pts_2d_x, pts_2d_y = np.meshgrid(x_coord_1d.points, y_coord_1d.points)
    if not x_coord_1d.has_bounds() or not y_coord_1d.has_bounds():
        bds_2d_x = None
        bds_2d_y = None
    else:
        bds_2d_x, bds_2d_y = expand_1d_x_and_y_bounds_to_2d_xy(
            x_coord_1d.bounds, y_coord_1d.bounds
        )

    # Make two new coords + return them.
    result = []
    for pts, bds, name in zip(
        (pts_2d_x, pts_2d_y), (bds_2d_x, bds_2d_y), ("longitude", "latitude")
    ):
        coord = AuxCoord(pts, bounds=bds, standard_name=name, units="degrees")
        result.append(coord)

    return result


def sample_2d_latlons(regional=False, rotated=False, transformed=False):
    """
    Construct small 2d cubes with 2d X and Y coordinates.

    This makes cubes with 'expanded' coordinates (4 bounds per cell), analagous
    to ORCA data.
    The coordinates are always geographical, so either it has a coord system
    or they are "true" lats + lons.
    ( At present, they are always latitudes and longitudes, but maybe in a
    rotated system. )
    The results always have fully contiguous bounds.

    Kwargs:
    * regional (bool):
        If False (default), results cover the whole globe, and there is
        implicit connectivity between rhs + lhs of the array.
        If True, coverage is regional and edges do not connect.
    * rotated (bool):
        If False, X and Y coordinates are true-latitudes and longitudes, with
        an implicit coordinate system (i.e. None).
        If True, the X and Y coordinates are lats+lons in a selected
        rotated-latlon coordinate system.
    * transformed (bool):
        Build coords from rotated coords as for 'rotated', but then replace
        their values with the equivalent "true" lats + lons, and no
        coord-system (defaults to true-latlon).
        In this case, the X and Y coords are no longer 'meshgrid' style,
        i.e. the points + bounds values vary in *both* dimensions.

    .. note::

        'transformed' is an alternative to 'rotated' :  when 'transformed' is
        set, then 'rotated' has no effect.

    .. Some sample results printouts ::

        >>> print(sample_2d_latlons())
        test_data / (unknown)               (-- : 5; -- : 6)
             Auxiliary coordinates:
                  latitude                      x       x
                  longitude                     x       x
        >>>
        >>> print(sample_2d_latlons().coord(axis='x')[0, :2])
        AuxCoord(array([ 37.5 ,  93.75]),
                 bounds=array([[   0.   ,   65.625,   65.625,    0.   ],
                               [  65.625,  121.875,  121.875,   65.625]]),
                 standard_name='longitude', units=Unit('degrees'))
        >>> print(np.round(sample_2d_latlons().coord(axis='x').points, 3))
        [[  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]]
        >>> print(np.round(sample_2d_latlons().coord(axis='y').points, 3))
        [[-85.  -85.  -85.  -85.  -85.  -85. ]
         [-47.5 -47.5 -47.5 -47.5 -47.5 -47.5]
         [-10.  -10.  -10.  -10.  -10.  -10. ]
         [ 27.5  27.5  27.5  27.5  27.5  27.5]
         [ 65.   65.   65.   65.   65.   65. ]]


        >>> print(np.round(
            sample_2d_latlons(rotated=True).coord(axis='x').points, 3))
        [[  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]
         [  37.5    93.75  150.    206.25  262.5   318.75]]
        >>> print(sample_2d_latlons(rotated=True).coord(axis='y').coord_system)
        RotatedGeogCS(75.0, 120.0)


        >>> print(
            sample_2d_latlons(transformed=True).coord(axis='y').coord_system)
        None
        >>> print(np.round(
            sample_2d_latlons(transformed=True).coord(axis='x').points, 3))
        [[ -50.718  -40.983  -46.74   -71.938  -79.293  -70.146]
         [ -29.867   17.606   77.936  157.145 -141.037  -93.172]
         [ -23.139   31.007   87.699  148.322 -154.639 -100.505]
         [ -16.054   41.218   92.761  143.837 -164.738 -108.105]
         [  10.86    61.78   100.236  137.285  175.511 -135.446]]
        >>> print(np.round(
            sample_2d_latlons(transformed=True).coord(axis='y').points, 3))
        [[-70.796 -74.52  -79.048 -79.26  -74.839 -70.96 ]
         [-34.99  -46.352 -59.721 -60.34  -47.305 -35.499]
         [  1.976 -10.626 -22.859 -23.349 -11.595   1.37 ]
         [ 38.914  25.531  14.312  13.893  24.585  38.215]
         [ 74.197  60.258  51.325  51.016  59.446  73.268]]
        >>>

    """

    def sample_cube(xargs, yargs):
        # Make a test cube with given latitude + longitude coordinates.
        # xargs/yargs are args for np.linspace (start, stop, N), to make the X
        # and Y coordinate points.
        x0, x1, nx = xargs
        y0, y1, ny = yargs
        # Data has cycling values, staggered a bit in successive rows.
        data = np.zeros((ny, nx))
        data.flat[:] = np.arange(ny * nx) % (nx + 2)
        # Build a 2d cube with longitude + latitude coordinates.
        cube = Cube(data, long_name="test_data")
        x_pts = np.linspace(x0, x1, nx, endpoint=True)
        y_pts = np.linspace(y0, y1, ny, endpoint=True)
        co_x = DimCoord(x_pts, standard_name="longitude", units="degrees")
        co_y = DimCoord(y_pts, standard_name="latitude", units="degrees")
        cube.add_dim_coord(co_y, 0)
        cube.add_dim_coord(co_x, 1)
        return cube

    # Start by making a "normal" cube with separate 1-D X and Y coords.
    if regional:
        # Make a small regional cube.
        cube = sample_cube(xargs=(150.0, 243.75, 6), yargs=(-10.0, 40.0, 5))

        # Add contiguous bounds.
        for ax in ("x", "y"):
            cube.coord(axis=ax).guess_bounds()
    else:
        # Global data, but at a drastically reduced resolution.
        cube = sample_cube(xargs=(37.5, 318.75, 6), yargs=(-85.0, 65.0, 5))

        # Make contiguous bounds and adjust outer edges to ensure it is global.
        for name in ("longitude", "latitude"):
            coord = cube.coord(name)
            coord.guess_bounds()
            bds = coord.bounds.copy()
            # Make bounds global, by fixing lowest and uppermost values.
            if name == "longitude":
                bds[0, 0] = 0.0
                bds[-1, 1] = 360.0
            else:
                bds[0, 0] = -90.0
                bds[-1, 1] = 90.0
            coord.bounds = bds

    # Now convert the 1-d coords to 2-d equivalents.
    # Get original 1-d coords.
    co_1d_x, co_1d_y = [cube.coord(axis=ax).copy() for ax in ("x", "y")]

    # Calculate 2-d equivalents.
    co_2d_x, co_2d_y = grid_coords_2d_from_1d(co_1d_x, co_1d_y)

    # Remove the old grid coords.
    for coord in (co_1d_x, co_1d_y):
        cube.remove_coord(coord)

    # Add the new grid coords.
    for coord in (co_2d_x, co_2d_y):
        cube.add_aux_coord(coord, (0, 1))

    if transformed or rotated:
        # Put the cube locations into a rotated coord system.
        pole_lat, pole_lon = 75.0, 120.0
        if transformed:
            # Reproject coordinate values from rotated to true lat-lons.
            co_x, co_y = [cube.coord(axis=ax) for ax in ("x", "y")]
            # Unrotate points.
            lons, lats = co_x.points, co_y.points
            lons, lats = unrotate_pole(lons, lats, pole_lon, pole_lat)
            co_x.points, co_y.points = lons, lats
            # Unrotate bounds.
            lons, lats = co_x.bounds, co_y.bounds
            # Note: save the shape, flatten + then re-apply the shape, because
            # "unrotate_pole" uses "cartopy.crs.CRS.transform_points", which
            # only works on arrays of 1 or 2 dimensions.
            shape = lons.shape
            lons, lats = unrotate_pole(
                lons.flatten(), lats.flatten(), pole_lon, pole_lat
            )
            co_x.bounds, co_y.bounds = lons.reshape(shape), lats.reshape(shape)
        else:
            # "Just" rotate operation : add a coord-system to each coord.
            cs = RotatedGeogCS(pole_lat, pole_lon)
            for coord in cube.coords():
                coord.coord_system = cs

    return cube


def make_bounds_discontiguous_at_point(cube, at_iy, at_ix, in_y=False):
    """
    Meddle with the XY grid bounds of a 2D cube to make the grid discontiguous.

    Changes the points and bounds of a single gridcell, so that it becomes
    discontinuous with an adjacent gridcell : either the one to its right, or
    the one above it (if 'in_y' is True).

    Also masks the cube data at the given point.

    The cube must be 2-dimensional and have bounded 2d 'x' and 'y' coordinates.

    """
    x_coord = cube.coord(axis="x")
    y_coord = cube.coord(axis="y")
    assert x_coord.shape == y_coord.shape
    assert (
        coord.bounds.ndim == 3 and coord.shape[-1] == 4
        for coord in (x_coord, y_coord)
    )

    # For both X and Y coord, move points + bounds to create a discontinuity.
    def adjust_coord(coord):
        pts, bds = coord.points, coord.bounds
        # Fetch the 4 bounds (bottom-left, bottom-right, top-right, top-left)
        bds_bl, bds_br, bds_tr, bds_tl = bds[at_iy, at_ix]
        if not in_y:
            # Make a discontinuity "at" (iy, ix), by moving the right-hand edge
            # of the cell to the midpoint of the existing left+right bounds.
            new_bds_br = 0.5 * (bds_bl + bds_br)
            new_bds_tr = 0.5 * (bds_tl + bds_tr)
            bds_br, bds_tr = new_bds_br, new_bds_tr
        else:
            # Same but in the 'grid y direction' :
            # Make a discontinuity "at" (iy, ix), by moving the **top** edge of
            # the cell to the midpoint of the existing **top+bottom** bounds.
            new_bds_tl = 0.5 * (bds_bl + bds_tl)
            new_bds_tr = 0.5 * (bds_br + bds_tr)
            bds_tl, bds_tr = new_bds_tl, new_bds_tr

        # Write in the new bounds (all 4 corners).
        bds[at_iy, at_ix] = [bds_bl, bds_br, bds_tr, bds_tl]
        # Also reset the cell midpoint to the middle of the 4 new corners,
        # in case having a midpoint outside the corners might cause a problem.
        new_pt = 0.25 * sum([bds_bl, bds_br, bds_tr, bds_tl])
        pts[at_iy, at_ix] = new_pt
        # Write back the coord points+bounds (can only assign whole arrays).
        coord.points, coord.bounds = pts, bds

    adjust_coord(x_coord)
    adjust_coord(y_coord)

    # Check which dimensions are spanned by each coordinate.
    for coord in (x_coord, y_coord):
        span = set(cube.coord_dims(coord))
        if not span:
            msg = "The coordinate {!r} doesn't span a data dimension."
            raise ValueError(msg.format(coord.name()))

        masked_data = ma.masked_array(cube.data)
        masked_data[at_iy, at_ix] = ma.masked

    cube.data = masked_data



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.fileformats.pp.save_pairs_from_cube` function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import iris.tests.stock as stock

from iris.fileformats.pp import save_pairs_from_cube


class TestSaveFields(tests.IrisTest):
    def setUp(self):
        self.cube = stock.realistic_3d()

    def test_cube_only(self):
        slices_and_fields = save_pairs_from_cube(self.cube)
        for aslice, field in slices_and_fields:
            self.assertEqual(aslice.shape, (9, 11))
            self.assertEqual(field.lbcode, 101)

    def test_field_coords(self):
        slices_and_fields = save_pairs_from_cube(
            self.cube, field_coords=["grid_longitude", "grid_latitude"]
        )
        for aslice, field in slices_and_fields:
            self.assertEqual(aslice.shape, (11, 9))
            self.assertEqual(field.lbcode, 101)

    def test_lazy_data(self):
        cube = self.cube.copy()
        # "Rebase" the cube onto a lazy version of its data.
        cube.data = cube.lazy_data()
        # Check that lazy data is preserved in save-pairs generation.
        slices_and_fields = save_pairs_from_cube(cube)
        for aslice, _ in slices_and_fields:
            self.assertTrue(aslice.has_lazy_data())

    def test_default_bmdi(self):
        slices_and_fields = save_pairs_from_cube(self.cube)
        _, field = next(slices_and_fields)
        self.assertEqual(field.bmdi, -1e30)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Unit tests for the :func:`iris.experimental.stratify.relevel` function.

"""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from functools import partial

import numpy as np
from numpy.testing import assert_array_equal

from iris.coords import AuxCoord, DimCoord
import iris.tests.stock as stock

try:
    import stratify
    from iris.experimental.stratify import relevel
except ImportError:
    stratify = None


@tests.skip_stratify
class Test(tests.IrisTest):
    def setUp(self):
        cube = stock.simple_3d()[:, :1, :1]
        #: The data from which to get the levels.
        self.src_levels = cube.copy()
        #: The data to interpolate.
        self.cube = cube.copy()
        self.cube.rename("foobar")
        self.cube *= 10
        self.coord = self.src_levels.coord("wibble")
        self.axes = (self.coord, self.coord.name(), None, 0)

    def test_broadcast_fail_src_levels(self):
        emsg = "Cannot broadcast the cube and src_levels"
        data = np.arange(60).reshape(3, 4, 5)
        with self.assertRaisesRegex(ValueError, emsg):
            relevel(self.cube, AuxCoord(data), [1, 2, 3])

    def test_broadcast_fail_tgt_levels(self):
        emsg = "Cannot broadcast the cube and tgt_levels"
        data = np.arange(60).reshape(3, 4, 5)
        with self.assertRaisesRegex(ValueError, emsg):
            relevel(self.cube, self.coord, data)

    def test_standard_input(self):
        for axis in self.axes:
            result = relevel(
                self.cube, self.src_levels, [-1, 0, 5.5], axis=axis
            )
            assert_array_equal(
                result.data.flatten(), np.array([np.nan, 0, 55])
            )
            expected = DimCoord([-1, 0, 5.5], units=1, long_name="thingness")
            self.assertEqual(expected, result.coord("thingness"))

    def test_non_monotonic(self):
        for axis in self.axes:
            result = relevel(self.cube, self.src_levels, [2, 3, 2], axis=axis)
            assert_array_equal(
                result.data.flatten(), np.array([20, 30, np.nan])
            )
            expected = AuxCoord([2, 3, 2], units=1, long_name="thingness")
            self.assertEqual(result.coord("thingness"), expected)

    def test_static_level(self):
        for axis in self.axes:
            result = relevel(self.cube, self.src_levels, [2, 2], axis=axis)
            assert_array_equal(result.data.flatten(), np.array([20, 20]))

    def test_coord_input(self):
        source = AuxCoord(self.src_levels.data)
        metadata = self.src_levels.metadata._asdict()
        metadata["coord_system"] = None
        metadata["climatological"] = None
        source.metadata = metadata

        for axis in self.axes:
            result = relevel(self.cube, source, [0, 12, 13], axis=axis)
            self.assertEqual(result.shape, (3, 1, 1))
            assert_array_equal(result.data.flatten(), [0, 120, np.nan])

    def test_custom_interpolator(self):
        interpolator = partial(stratify.interpolate, interpolation="nearest")

        for axis in self.axes:
            result = relevel(
                self.cube,
                self.src_levels,
                [-1, 0, 6.5],
                axis=axis,
                interpolator=interpolator,
            )
            assert_array_equal(
                result.data.flatten(), np.array([np.nan, 0, 120])
            )

    def test_multi_dim_target_levels(self):
        interpolator = partial(
            stratify.interpolate,
            interpolation="linear",
            extrapolation="linear",
        )

        for axis in self.axes:
            result = relevel(
                self.cube,
                self.src_levels,
                self.src_levels.data,
                axis=axis,
                interpolator=interpolator,
            )
            assert_array_equal(result.data.flatten(), np.array([0, 120]))
            self.assertCML(result)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Integration tests for experimental regridding."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import cartopy.crs as ccrs
from cf_units import Unit
import numpy as np

import iris
import iris.aux_factory
from iris.coord_systems import GeogCS
from iris.tests.stock import global_pp
from iris.experimental.regrid import (
    ProjectedUnstructuredLinear,
    ProjectedUnstructuredNearest,
)


@tests.skip_data
class TestProjectedUnstructured(tests.IrisTest):
    def setUp(self):
        path = tests.get_data_path(
            ("NetCDF", "unstructured_grid", "theta_nodal_xios.nc")
        )
        self.src = iris.load_cube(path, "Potential Temperature")

        src_lat = self.src.coord("latitude")
        src_lon = self.src.coord("longitude")
        src_lat.coord_system = src_lon.coord_system = GeogCS(6370000)
        src_lat.convert_units(Unit("degrees"))
        src_lon.convert_units(Unit("degrees"))

        self.global_grid = global_pp()

    def test_nearest(self):
        res = self.src.regrid(self.global_grid, ProjectedUnstructuredNearest())
        self.assertArrayShapeStats(
            res, (1, 6, 73, 96), 315.8913582, 11.00063922733, rtol=1e-8
        )
        self.assertArrayShapeStats(
            res[:, 0], (1, 73, 96), 299.99993826, 3.9226378869e-5
        )

    def test_nearest_sinusoidal(self):
        crs = ccrs.Sinusoidal()
        res = self.src.regrid(
            self.global_grid, ProjectedUnstructuredNearest(crs)
        )
        self.assertArrayShapeStats(
            res, (1, 6, 73, 96), 315.891358296, 11.000639227, rtol=1e-8
        )
        self.assertArrayShapeStats(
            res[:, 0], (1, 73, 96), 299.99993826, 3.9223839688e-5
        )

    def test_nearest_gnomonic_uk_domain(self):
        crs = ccrs.Gnomonic(central_latitude=60.0)
        uk_grid = self.global_grid.intersection(
            longitude=(-20, 20), latitude=(40, 80)
        )
        res = self.src.regrid(uk_grid, ProjectedUnstructuredNearest(crs))

        self.assertArrayShapeStats(
            res, (1, 6, 17, 11), 315.8873266, 11.0006664668, rtol=1e-8
        )
        self.assertArrayShapeStats(
            res[:, 0], (1, 17, 11), 299.99993826, 4.1356150388e-5
        )
        expected_subset = np.array(
            [
                [318.936829, 318.936829, 318.936829],
                [318.936829, 318.936829, 318.936829],
                [318.935163, 318.935163, 318.935163],
            ]
        )
        self.assertArrayAlmostEqual(
            expected_subset, res.data[0, 3, 5:8, 4:7].data
        )

    def test_nearest_aux_factories(self):
        src = self.src

        (xy_dim_len,) = src.coord(axis="X").shape
        (z_dim_len,) = src.coord("levels").shape

        src.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(z_dim_len) + 40, long_name="level_height", units="m"
            ),
            1,
        )
        src.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(z_dim_len) + 50, long_name="sigma", units="1"
            ),
            1,
        )
        src.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(xy_dim_len) + 100,
                long_name="surface_altitude",
                units="m",
            ),
            2,
        )
        src.add_aux_factory(
            iris.aux_factory.HybridHeightFactory(
                delta=src.coord("level_height"),
                sigma=src.coord("sigma"),
                orography=src.coord("surface_altitude"),
            )
        )
        res = src.regrid(self.global_grid, ProjectedUnstructuredNearest())

        self.assertArrayShapeStats(
            res, (1, 6, 73, 96), 315.8913582, 11.000639227334, rtol=1e-8
        )
        self.assertArrayShapeStats(
            res[:, 0], (1, 73, 96), 299.99993826, 3.9226378869e-5
        )
        self.assertEqual(res.coord("altitude").shape, (6, 73, 96))

    def test_linear_sinusoidal(self):
        res = self.src.regrid(self.global_grid, ProjectedUnstructuredLinear())
        self.assertArrayShapeStats(
            res, (1, 6, 73, 96), 315.8914839, 11.0006338412, rtol=1e-8
        )
        self.assertArrayShapeStats(
            res[:, 0], (1, 73, 96), 299.99993826, 3.775024069e-5
        )
        expected_subset = np.array(
            [
                [299.999987, 299.999996, 299.999999],
                [299.999984, 299.999986, 299.999988],
                [299.999973, 299.999977, 299.999982],
            ]
        )
        self.assertArrayAlmostEqual(
            expected_subset, res.data[0, 0, 20:23, 40:43].data
        )


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.plot._check_geostationary_coords_and_convert
function."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from unittest.mock import Mock

import numpy as np

from cartopy.crs import Geostationary, NearsidePerspective
from iris.plot import _check_geostationary_coords_and_convert


class Test__check_geostationary_coords_and_convert(tests.IrisTest):
    def setUp(self):
        geostationary_altitude = 35785831.0
        # proj4_params is the one attribute of the Geostationary class that
        # is needed for the function.
        self.proj4_params = {"h": geostationary_altitude}

        # Simulate the maximum-dimension array that could be processed.
        a = np.linspace(0, 2, 6)
        b = np.linspace(2, 3, 5)
        self.x_original, self.y_original = np.meshgrid(a, b)

        # Expected arrays if conversion takes place.
        self.x_converted, self.y_converted = (
            i * geostationary_altitude
            for i in (self.x_original, self.y_original)
        )

    def _test(self, geostationary=True):
        # Re-usable test for when Geostationary is present OR absent.
        if geostationary:
            # A Geostationary projection WILL be processed.
            projection_spec = Geostationary
            target_tuple = (self.x_converted, self.y_converted)
        else:
            # A non-Geostationary projection WILL NOT be processed.
            projection_spec = NearsidePerspective
            target_tuple = (self.x_original, self.y_original)

        projection = Mock(spec=projection_spec)
        projection.proj4_params = self.proj4_params
        # Projection is looked for within a dictionary called kwargs.
        kwargs = {"transform": projection}

        x, y = _check_geostationary_coords_and_convert(
            self.x_original, self.y_original, kwargs
        )
        self.assertArrayEqual((x, y), target_tuple)

    def test_geostationary_present(self):
        self._test(geostationary=True)

    def test_geostationary_absent(self):
        self._test(geostationary=False)



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Test the iteration of cubes in step.

"""

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import itertools
import operator
import random
import warnings

import numpy as np

import iris
import iris.analysis
import iris.iterate
import iris.tests.stock
from functools import reduce


@tests.skip_data
class TestIterateFunctions(tests.IrisTest):
    def setUp(self):
        self.cube_a = iris.tests.stock.realistic_4d()[0, 0:5, 0:10, 0:12]
        self.cube_b = iris.tests.stock.realistic_4d()[1, 0:5, 0:10, 0:12]
        self.coord_names = ["grid_latitude", "grid_longitude"]

        # Modify elements of cube_b to introduce additional differences
        self.cube_b.attributes["source"] = "Iris iterate test case"
        self.cube_b.add_aux_coord(iris.coords.AuxCoord(23, long_name="other"))

    def test_izip_no_args(self):
        with self.assertRaises(TypeError):
            iris.iterate.izip()
        with self.assertRaises(TypeError):
            iris.iterate.izip(coords=self.coord_names)
        with self.assertRaises(TypeError):
            iris.iterate.izip(coords=self.coord_names, ordered=False)

    def test_izip_input_collections(self):
        # Should work with one or more cubes as args
        iris.iterate.izip(self.cube_a, coords=self.coord_names)
        iris.iterate.izip(self.cube_a, self.cube_a, coords=self.coord_names)
        iris.iterate.izip(self.cube_a, self.cube_b, coords=self.coord_names)
        iris.iterate.izip(
            self.cube_a, self.cube_b, self.cube_a, coords=self.coord_names
        )
        # Check unpacked collections
        cubes = [self.cube_a] * 10
        iris.iterate.izip(*cubes, coords=self.coord_names)
        cubes = tuple(cubes)
        iris.iterate.izip(*cubes, coords=self.coord_names)

    def test_izip_returns_iterable(self):
        try:
            # Raises an exception if arg is not iterable
            iter(iris.iterate.izip(self.cube_a, coords=self.coord_names))
        except TypeError:
            self.fail("iris.iterate.izip is not returning an iterable")

    def test_izip_unequal_slice_coords(self):
        # Create a cube with grid_latitude and grid_longitude coords
        # that differ in size from cube_a's
        other_cube = self.cube_a[0, 0:3, 0:3]
        nslices = self.cube_a.shape[0]
        i = 0
        for slice_a, slice_other in iris.iterate.izip(
            self.cube_a, other_cube, coords=self.coord_names
        ):
            slice_a_truth = self.cube_a[i, :, :]
            slice_other_truth = other_cube
            self.assertEqual(slice_a_truth, slice_a)
            self.assertEqual(slice_other_truth, slice_other)
            i += 1
        self.assertEqual(i, nslices)
        # Attempting to iterate over these incompatible coords should
        # raise an exception
        with self.assertRaises(ValueError):
            iris.iterate.izip(self.cube_a, other_cube)

    def test_izip_missing_slice_coords(self):
        # Remove latitude coordinate from one of the cubes
        other_cube = self.cube_b.copy()
        other_cube.remove_coord("grid_latitude")
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            iris.iterate.izip(self.cube_a, other_cube, coords=self.coord_names)
        # Create a cube with latitude and longitude rather than grid_latitude
        # and grid_longitude
        self.cube_b.coord("grid_latitude").rename("latitude")
        self.cube_b.coord("grid_longitude").rename("longitude")
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            iris.iterate.izip(
                self.cube_a, self.cube_b, coords=self.coord_names
            )

    def test_izip_onecube_no_coords(self):
        # Should do the same as slices() but bearing in mind izip.next()
        # returns a tuple of cubes

        # Reduce the size to speed things up - we're going to iterate
        # over *every* data point.
        self.cube_b = self.cube_b[:2, :4, :3]

        # Empty list as coords
        slice_iterator = self.cube_b.slices([])
        zip_iterator = iris.iterate.izip(self.cube_b, coords=[])
        for cube_slice in slice_iterator:
            # First element of tuple: (extractedcube, )
            zip_slice = next(zip_iterator)[0]
            self.assertEqual(cube_slice, zip_slice)
        with self.assertRaises(StopIteration):
            next(zip_iterator)  # Should raise exception if we continue try to
            # to iterate

    def test_izip_onecube_lat_lon(self):
        # Two coords
        slice_iterator = self.cube_b.slices(self.coord_names)
        zip_iterator = iris.iterate.izip(self.cube_b, coords=self.coord_names)
        for cube_slice in slice_iterator:
            # First element of tuple: (extractedcube, )
            zip_slice = next(zip_iterator)[0]
            self.assertEqual(cube_slice, zip_slice)
        with self.assertRaises(StopIteration):
            next(zip_iterator)  # Should raise exception if we continue to try
            # to iterate

    def test_izip_onecube_lat(self):
        # One coord
        slice_iterator = self.cube_b.slices("grid_latitude")
        zip_iterator = iris.iterate.izip(self.cube_b, coords="grid_latitude")
        for cube_slice in slice_iterator:
            # First element of tuple: (extractedcube, )
            zip_slice = next(zip_iterator)[0]
            self.assertEqual(cube_slice, zip_slice)
        with self.assertRaises(StopIteration):
            next(zip_iterator)  # Should raise exception if we continue to try
            # to iterate

    def test_izip_onecube_height_lat_long(self):
        # All coords
        slice_iterator = self.cube_b.slices(
            ["level_height", "grid_latitude", "grid_longitude"]
        )
        zip_iterator = iris.iterate.izip(
            self.cube_b,
            coords=["level_height", "grid_latitude", "grid_longitude"],
        )
        for cube_slice in slice_iterator:
            # First element of tuple: (extractedcube, )
            zip_slice = next(zip_iterator)[0]
            self.assertEqual(cube_slice, zip_slice)
        with self.assertRaises(StopIteration):
            next(zip_iterator)  # Should raise exception if we continue to try
            # to iterate

    def test_izip_same_cube_lat_lon(self):
        nslices = self.cube_b.shape[0]
        slice_iterator = self.cube_b.slices(self.coord_names)
        count = 0
        for slice_first, slice_second in iris.iterate.izip(
            self.cube_b, self.cube_b, coords=self.coord_names
        ):
            self.assertEqual(slice_first, slice_second)  # Equal to each other
            self.assertEqual(
                slice_first, next(slice_iterator)
            )  # Equal to the truth (from slice())
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_same_cube_lat(self):
        nslices = (
            self.cube_a.shape[0] * self.cube_a.shape[2]
        )  # Calc product of dimensions
        # excluding the latitude
        # (2nd data dim)
        slice_iterator = self.cube_a.slices("grid_latitude")
        count = 0
        for slice_first, slice_second in iris.iterate.izip(
            self.cube_a, self.cube_a, coords=["grid_latitude"]
        ):
            self.assertEqual(slice_first, slice_second)
            self.assertEqual(
                slice_first, next(slice_iterator)
            )  # Equal to the truth (from slice())
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_same_cube_no_coords(self):
        # Reduce the size to speed things up - we're going to iterate
        # over *every* data point.
        self.cube_b = self.cube_b[:2, :4, :3]

        nslices = reduce(operator.mul, self.cube_b.shape)
        slice_iterator = self.cube_b.slices([])
        count = 0
        for slice_first, slice_second in iris.iterate.izip(
            self.cube_b, self.cube_b, coords=[]
        ):
            self.assertEqual(slice_first, slice_second)
            self.assertEqual(
                slice_first, next(slice_iterator)
            )  # Equal to the truth (from slice())
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_subcube_of_same(self):
        for _ in range(3):
            super_cube = self.cube_a
            # Random int to pick coord value to calc subcube
            k = random.randint(0, super_cube.shape[0] - 1)
            sub_cube = super_cube[k, :, :]
            super_slice_iterator = super_cube.slices(self.coord_names)
            j = 0
            for super_slice, sub_slice in iris.iterate.izip(
                super_cube, sub_cube, coords=self.coord_names
            ):
                self.assertEqual(
                    sub_slice, sub_cube
                )  # This cube should not change
                # as lat and long are the only
                # data dimensions in this cube)
                self.assertEqual(super_slice, next(super_slice_iterator))
                if j == k:
                    self.assertEqual(super_slice, sub_slice)
                else:
                    self.assertNotEqual(super_slice, sub_slice)
                j += 1
            nslices = super_cube.shape[0]
            self.assertEqual(j, nslices)

    def test_izip_same_dims(self):
        # Check single coords slice
        nslices = reduce(operator.mul, self.cube_a.shape[1:])
        nslices_to_check = (
            20  # This is only approximate as we use random to select slices
        )
        # Fraction of slices to check
        check_eq_probability = max(0.0, min(1.0, nslices_to_check / nslices))

        ij_iterator = np.ndindex(self.cube_a.shape[1], self.cube_a.shape[2])
        count = 0
        for slice_a, slice_b in iris.iterate.izip(
            self.cube_a, self.cube_b, coords="level_height"
        ):
            i, j = next(ij_iterator)
            if random.random() < check_eq_probability:  # Check these slices
                slice_a_truth = self.cube_a[:, i, j]
                slice_b_truth = self.cube_b[:, i, j]
                self.assertEqual(slice_a_truth, slice_a)
                self.assertEqual(slice_b_truth, slice_b)
            count += 1
        self.assertEqual(count, nslices)
        # Two coords
        nslices = self.cube_a.shape[0]
        i_iterator = iter(range(self.cube_a.shape[0]))
        count = 0
        for slice_a, slice_b in iris.iterate.izip(
            self.cube_a, self.cube_b, coords=self.coord_names
        ):
            i = next(i_iterator)
            slice_a_truth = self.cube_a[i, :, :]
            slice_b_truth = self.cube_b[i, :, :]
            self.assertEqual(slice_a_truth, slice_a)
            self.assertEqual(slice_b_truth, slice_b)
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_extra_dim(self):
        big_cube = self.cube_a
        # Remove first data dimension and associated coords
        little_cube = self.cube_b.copy()
        for factory in little_cube.aux_factories:
            little_cube.remove_aux_factory(factory)
        little_cube = little_cube[0]
        little_cube.remove_coord("model_level_number")
        little_cube.remove_coord("level_height")
        little_cube.remove_coord("sigma")
        # little_slice should remain the same as there are no other data dimensions
        little_slice_truth = little_cube
        i = 0
        for big_slice, little_slice in iris.iterate.izip(
            big_cube, little_cube, coords=self.coord_names
        ):
            big_slice_truth = big_cube[i, :, :]
            self.assertEqual(little_slice_truth, little_slice)
            self.assertEqual(big_slice_truth, big_slice)
            i += 1
        nslices = big_cube.shape[0]
        self.assertEqual(nslices, i)

        # Leave middle coord but move it from a data dimension to a scalar coord by slicing
        little_cube = self.cube_b[:, 0, :]

        # Now remove associated coord
        little_cube.remove_coord("grid_latitude")
        # Check we raise an exception if we request coords one of the cubes doesn't have
        with self.assertRaises(iris.exceptions.CoordinateNotFoundError):
            iris.iterate.izip(big_cube, little_cube, coords=self.coord_names)

        # little_slice should remain the same as there are no other data dimensions
        little_slice_truth = little_cube
        i = 0
        for big_slice, little_slice in iris.iterate.izip(
            big_cube,
            little_cube,
            coords=["model_level_number", "grid_longitude"],
        ):
            big_slice_truth = big_cube[:, i, :]
            self.assertEqual(little_slice_truth, little_slice)
            self.assertEqual(big_slice_truth, big_slice)
            i += 1
        nslices = big_cube.shape[1]
        self.assertEqual(nslices, i)

        # Take a random slice reducing it to a 1d cube
        p = random.randint(0, self.cube_b.shape[0] - 1)
        q = random.randint(0, self.cube_b.shape[2] - 1)
        little_cube = self.cube_b[p, :, q]
        nslices = big_cube.shape[0] * big_cube.shape[2]
        nslices_to_check = (
            20  # This is only approximate as we use random to select slices
        )
        # Fraction of slices to check
        check_eq_probability = max(0.0, min(1.0, nslices_to_check / nslices))
        ij_iterator = np.ndindex(big_cube.shape[0], big_cube.shape[2])
        count = 0
        for big_slice, little_slice in iris.iterate.izip(
            big_cube, little_cube, coords="grid_latitude"
        ):
            i, j = next(ij_iterator)
            if random.random() < check_eq_probability:
                big_slice_truth = big_cube[i, :, j]
                little_slice_truth = (
                    little_cube  # Just 1d so slice is entire cube
                )
                self.assertEqual(little_slice_truth, little_slice)
                self.assertEqual(big_slice_truth, big_slice)
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_different_shaped_coords(self):
        other = self.cube_b[0:-1]
        # Different 'z' coord shape - expect a ValueError
        with self.assertRaises(ValueError):
            iris.iterate.izip(self.cube_a, other, coords=self.coord_names)

    def test_izip_different_valued_coords(self):
        # Change a value in one of the coord points arrays so they are no longer identical
        new_points = self.cube_b.coord("model_level_number").points.copy()
        new_points[0] = 0
        self.cube_b.coord("model_level_number").points = new_points
        # slice coords
        latitude = self.cube_b.coord("grid_latitude")
        longitude = self.cube_b.coord("grid_longitude")
        # Same coord metadata and shape, but different values - check it produces a warning
        with warnings.catch_warnings():
            warnings.simplefilter(
                "error"
            )  # Cause all warnings to raise Exceptions
            with self.assertRaises(UserWarning):
                iris.iterate.izip(
                    self.cube_a, self.cube_b, coords=self.coord_names
                )
            # Call with coordinates, rather than names
            with self.assertRaises(UserWarning):
                iris.iterate.izip(
                    self.cube_a, self.cube_b, coords=[latitude, longitude]
                )
        # Check it still iterates through as expected
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            nslices = self.cube_a.shape[0]
            i = 0
            for slice_a, slice_b in iris.iterate.izip(
                self.cube_a, self.cube_b, coords=self.coord_names
            ):
                slice_a_truth = self.cube_a[i, :, :]
                slice_b_truth = self.cube_b[i, :, :]
                self.assertEqual(slice_a_truth, slice_a)
                self.assertEqual(slice_b_truth, slice_b)
                self.assertNotEqual(slice_b, None)
                i += 1
            self.assertEqual(i, nslices)
            # Call with coordinate instances rather than coord names
            i = 0
            for slice_a, slice_b in iris.iterate.izip(
                self.cube_a, self.cube_b, coords=[latitude, longitude]
            ):
                slice_a_truth = self.cube_a[i, :, :]
                slice_b_truth = self.cube_b[i, :, :]
                self.assertEqual(slice_a_truth, slice_a)
                self.assertEqual(slice_b_truth, slice_b)
                i += 1
            self.assertEqual(i, nslices)

    def test_izip_ordered(self):
        # Remove coordinate that spans grid_latitude and
        # grid_longitude dimensions as this will be common between
        # the resulting cubes but differ in shape
        self.cube_b.remove_coord("surface_altitude")
        cube = self.cube_b.copy()
        cube.transpose([0, 2, 1])  # switch order of lat and lon
        nslices = self.cube_b.shape[0]
        # Default behaviour: ordered = True
        i = 0
        for slice_b, cube_slice in iris.iterate.izip(
            self.cube_b, cube, coords=self.coord_names, ordered=True
        ):
            slice_b_truth = self.cube_b[i, :, :]
            cube_slice_truth = cube[i, :, :]
            # izip should transpose the slice to ensure order is [lat, lon]
            cube_slice_truth.transpose()
            self.assertEqual(slice_b_truth, slice_b)
            self.assertEqual(cube_slice_truth, cube_slice)
            i += 1
        self.assertEqual(i, nslices)
        # Alternative behaviour: ordered=False (retain original ordering)
        i = 0
        for slice_b, cube_slice in iris.iterate.izip(
            self.cube_b, cube, coords=self.coord_names, ordered=False
        ):
            slice_b_truth = self.cube_b[i, :, :]
            cube_slice_truth = cube[i, :, :]
            self.assertEqual(slice_b_truth, slice_b)
            self.assertEqual(cube_slice_truth, cube_slice)
            i += 1
        self.assertEqual(i, nslices)

    def test_izip_use_in_analysis(self):
        # Calculate mean, collapsing vertical dimension
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            vertical_mean = self.cube_b.collapsed(
                "model_level_number", iris.analysis.MEAN
            )
        nslices = self.cube_b.shape[0]
        i = 0
        for slice_b, mean_slice in iris.iterate.izip(
            self.cube_b, vertical_mean, coords=self.coord_names
        ):
            slice_b_truth = self.cube_b[i, :, :]
            self.assertEqual(slice_b_truth, slice_b)
            # Should return same cube in each iteration
            self.assertEqual(vertical_mean, mean_slice)
            i += 1
        self.assertEqual(i, nslices)

    def test_izip_nd_non_ortho(self):
        cube1 = iris.cube.Cube(np.zeros((5, 5, 5)))
        cube1.add_aux_coord(
            iris.coords.AuxCoord(np.arange(5), long_name="z"), [0]
        )
        cube1.add_aux_coord(
            iris.coords.AuxCoord(np.arange(25).reshape(5, 5), long_name="y"),
            [1, 2],
        )
        cube1.add_aux_coord(
            iris.coords.AuxCoord(np.arange(25).reshape(5, 5), long_name="x"),
            [1, 2],
        )
        cube2 = cube1.copy()

        # The two coords are not orthogonal so we cannot use them with izip
        with self.assertRaises(ValueError):
            iris.iterate.izip(cube1, cube2, coords=["y", "x"])

    def test_izip_nd_ortho(self):
        cube1 = iris.cube.Cube(np.zeros((5, 5, 5, 5, 5), dtype="f8"))
        cube1.add_dim_coord(
            iris.coords.DimCoord(
                np.arange(5, dtype="i8"), long_name="z", units="1"
            ),
            [0],
        )
        cube1.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(25, dtype="i8").reshape(5, 5),
                long_name="y",
                units="1",
            ),
            [1, 2],
        )
        cube1.add_aux_coord(
            iris.coords.AuxCoord(
                np.arange(25, dtype="i8").reshape(5, 5),
                long_name="x",
                units="1",
            ),
            [3, 4],
        )
        cube2 = cube1.copy()

        # The two coords are orthogonal so we can use them with izip
        it = iris.iterate.izip(cube1, cube2, coords=["y", "x"])
        cubes = list(np.array(list(it)).flatten())
        self.assertCML(cubes, ("iterate", "izip_nd_ortho.cml"))

    def _check_2d_slices(self):
        # Helper method to verify slices from izip match those from
        # cube.slices().
        slice_a_iterator = self.cube_a.slices(self.coord_names)
        slice_b_iterator = self.cube_b.slices(self.coord_names)
        nslices = self.cube_b.shape[0]
        count = 0
        for slice_a, slice_b in iris.iterate.izip(
            self.cube_a, self.cube_b, coords=self.coord_names
        ):
            self.assertEqual(slice_a, next(slice_a_iterator))
            self.assertEqual(slice_b, next(slice_b_iterator))
            count += 1
        self.assertEqual(count, nslices)

    def test_izip_extra_coords_step_dim(self):
        # Add extra different coords to cubes along the dimension we are
        # stepping through.
        coord_a = iris.coords.AuxCoord(
            np.arange(self.cube_a.shape[0]), long_name="another on a"
        )
        self.cube_a.add_aux_coord(coord_a, 0)
        coord_b = iris.coords.AuxCoord(
            np.arange(self.cube_b.shape[0]), long_name="another on b"
        )
        self.cube_b.add_aux_coord(coord_b, 0)
        # Check slices.
        self._check_2d_slices()

    def test_izip_extra_coords_slice_dim(self):
        # Add extra different coords to cubes along a dimension we are
        # not stepping through.
        coord_a = iris.coords.AuxCoord(
            np.arange(self.cube_a.shape[1]), long_name="another on a"
        )
        self.cube_a.add_aux_coord(coord_a, 1)
        coord_b = iris.coords.AuxCoord(
            np.arange(self.cube_b.shape[1]), long_name="another on b"
        )
        self.cube_b.add_aux_coord(coord_b, 1)
        self._check_2d_slices()

    def test_izip_extra_coords_both_slice_dims(self):
        # Add extra different coords to cubes along the dimensions we are
        # not stepping through.
        coord_a = iris.coords.AuxCoord(
            np.arange(self.cube_a.shape[1]), long_name="another on a"
        )
        self.cube_a.add_aux_coord(coord_a, 1)
        coord_b = iris.coords.AuxCoord(
            np.arange(self.cube_b.shape[2]), long_name="another on b"
        )
        self.cube_b.add_aux_coord(coord_b, 2)
        self._check_2d_slices()

    def test_izip_no_common_coords_on_step_dim(self):
        # Change metadata on all coords along the dimension we are
        # stepping through.
        self.cube_a.coord("model_level_number").rename("foo")
        self.cube_a.coord("sigma").rename("bar")
        self.cube_a.coord("level_height").rename("woof")
        # izip should step through them as a product.
        slice_a_iterator = self.cube_a.slices(self.coord_names)
        slice_b_iterator = self.cube_b.slices(self.coord_names)
        product_iterator = itertools.product(
            slice_a_iterator, slice_b_iterator
        )
        nslices = self.cube_a.shape[0] * self.cube_b.shape[0]
        count = 0
        for slice_a, slice_b in iris.iterate.izip(
            self.cube_a, self.cube_b, coords=self.coord_names
        ):
            expected_a, expected_b = next(product_iterator)
            self.assertEqual(slice_a, expected_a)
            self.assertEqual(slice_b, expected_b)
            count += 1
        self.assertEqual(count, nslices)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Test function :func:`iris.util.reverse`."""

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

import unittest

import iris
from iris.util import reverse
import numpy as np


class Test_array(tests.IrisTest):
    def test_simple_array(self):
        a = np.arange(12).reshape(3, 4)
        self.assertArrayEqual(a[::-1], reverse(a, 0))
        self.assertArrayEqual(a[::-1, ::-1], reverse(a, [0, 1]))
        self.assertArrayEqual(a[:, ::-1], reverse(a, 1))
        self.assertArrayEqual(a[:, ::-1], reverse(a, [1]))

        msg = "Reverse was expecting a single axis or a 1d array *"
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [])

        msg = "An axis value out of range for the number of dimensions *"
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, -1)
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, 10)
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [-1])
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [0, -1])

        msg = "To reverse an array, provide an int *"
        with self.assertRaisesRegex(TypeError, msg):
            reverse(a, "latitude")

    def test_single_array(self):
        a = np.arange(36).reshape(3, 4, 3)
        self.assertArrayEqual(a[::-1], reverse(a, 0))
        self.assertArrayEqual(a[::-1, ::-1], reverse(a, [0, 1]))
        self.assertArrayEqual(a[:, ::-1, ::-1], reverse(a, [1, 2]))
        self.assertArrayEqual(a[..., ::-1], reverse(a, 2))

        msg = "Reverse was expecting a single axis or a 1d array *"
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [])

        msg = "An axis value out of range for the number of dimensions *"
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, -1)
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, 10)
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [-1])
        with self.assertRaisesRegex(ValueError, msg):
            reverse(a, [0, -1])

        with self.assertRaisesRegex(
            TypeError, "To reverse an array, provide an int *"
        ):
            reverse(a, "latitude")


class Test_cube(tests.IrisTest):
    def setUp(self):
        # On this cube pair, the coordinates to perform operations on have
        # matching long names but the points array on one cube is reversed
        # with respect to that on the other.
        data = np.arange(12).reshape(3, 4)
        self.a1 = iris.coords.DimCoord([1, 2, 3], long_name="a")
        self.b1 = iris.coords.DimCoord([1, 2, 3, 4], long_name="b")
        a2 = iris.coords.DimCoord([3, 2, 1], long_name="a")
        b2 = iris.coords.DimCoord([4, 3, 2, 1], long_name="b")
        self.span = iris.coords.AuxCoord(
            np.arange(12).reshape(3, 4), long_name="spanning"
        )

        self.cube1 = iris.cube.Cube(
            data,
            dim_coords_and_dims=[(self.a1, 0), (self.b1, 1)],
            aux_coords_and_dims=[(self.span, (0, 1))],
        )

        self.cube2 = iris.cube.Cube(
            data, dim_coords_and_dims=[(a2, 0), (b2, 1)]
        )

    def test_cube_dim(self):
        cube1_reverse0 = reverse(self.cube1, 0)
        cube1_reverse1 = reverse(self.cube1, 1)
        cube1_reverse_both = reverse(self.cube1, (0, 1))

        self.assertArrayEqual(self.cube1.data[::-1], cube1_reverse0.data)
        self.assertArrayEqual(
            self.cube2.coord("a").points, cube1_reverse0.coord("a").points
        )
        self.assertArrayEqual(
            self.cube1.coord("b").points, cube1_reverse0.coord("b").points
        )

        self.assertArrayEqual(self.cube1.data[:, ::-1], cube1_reverse1.data)
        self.assertArrayEqual(
            self.cube1.coord("a").points, cube1_reverse1.coord("a").points
        )
        self.assertArrayEqual(
            self.cube2.coord("b").points, cube1_reverse1.coord("b").points
        )

        self.assertArrayEqual(
            self.cube1.data[::-1, ::-1], cube1_reverse_both.data
        )
        self.assertArrayEqual(
            self.cube2.coord("a").points, cube1_reverse_both.coord("a").points
        )
        self.assertArrayEqual(
            self.cube2.coord("b").points, cube1_reverse_both.coord("b").points
        )

    def test_cube_coord(self):
        cube1_reverse0 = reverse(self.cube1, self.a1)
        cube1_reverse1 = reverse(self.cube1, "b")
        cube1_reverse_both = reverse(self.cube1, (self.a1, self.b1))
        cube1_reverse_spanning = reverse(self.cube1, "spanning")

        self.assertArrayEqual(self.cube1.data[::-1], cube1_reverse0.data)
        self.assertArrayEqual(
            self.cube2.coord("a").points, cube1_reverse0.coord("a").points
        )
        self.assertArrayEqual(
            self.cube1.coord("b").points, cube1_reverse0.coord("b").points
        )

        self.assertArrayEqual(self.cube1.data[:, ::-1], cube1_reverse1.data)
        self.assertArrayEqual(
            self.cube1.coord("a").points, cube1_reverse1.coord("a").points
        )
        self.assertArrayEqual(
            self.cube2.coord("b").points, cube1_reverse1.coord("b").points
        )

        self.assertArrayEqual(
            self.cube1.data[::-1, ::-1], cube1_reverse_both.data
        )
        self.assertArrayEqual(
            self.cube2.coord("a").points, cube1_reverse_both.coord("a").points
        )
        self.assertArrayEqual(
            self.cube2.coord("b").points, cube1_reverse_both.coord("b").points
        )

        self.assertArrayEqual(
            self.cube1.data[::-1, ::-1], cube1_reverse_spanning.data
        )
        self.assertArrayEqual(
            self.cube2.coord("a").points,
            cube1_reverse_spanning.coord("a").points,
        )
        self.assertArrayEqual(
            self.cube2.coord("b").points,
            cube1_reverse_spanning.coord("b").points,
        )
        self.assertArrayEqual(
            self.span.points[::-1, ::-1],
            cube1_reverse_spanning.coord("spanning").points,
        )

        msg = "Expected to find exactly 1 latitude coordinate, but found none."
        with self.assertRaisesRegex(
            iris.exceptions.CoordinateNotFoundError, msg
        ):
            reverse(self.cube1, "latitude")

        msg = "Reverse was expecting a single axis or a 1d array *"
        with self.assertRaisesRegex(ValueError, msg):
            reverse(self.cube1, [])

        msg = (
            "coords_or_dims must be int, str, coordinate or sequence of "
            "these.  Got cube."
        )
        with self.assertRaisesRegex(TypeError, msg):
            reverse(self.cube1, self.cube1)

        msg = (
            "coords_or_dims must be int, str, coordinate or sequence of "
            "these."
        )
        with self.assertRaisesRegex(TypeError, msg):
            reverse(self.cube1, 3.0)


if __name__ == "__main__":
    unittest.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# import iris tests first so that some things can be initialised before
# importing anything else
import iris.tests as tests

import copy
import datetime
import unittest

import cf_units
import cftime
import matplotlib.units
import numpy as np


# Importing pandas has the side-effect of messing with the formatters
# used by matplotlib for handling dates.
default_units_registry = copy.copy(matplotlib.units.registry)
try:
    import pandas
except ImportError:
    # Disable all these tests if pandas is not installed.
    pandas = None
matplotlib.units.registry = default_units_registry

skip_pandas = unittest.skipIf(
    pandas is None, 'Test(s) require "pandas", ' "which is not available."
)

if pandas is not None:
    from iris.coords import DimCoord
    from iris.cube import Cube
    import iris.pandas


@skip_pandas
class TestAsSeries(tests.IrisTest):
    """Test conversion of 1D cubes to Pandas using as_series()"""

    def test_no_dim_coord(self):
        cube = Cube(np.array([0, 1, 2, 3, 4]), long_name="foo")
        series = iris.pandas.as_series(cube)
        expected_index = np.array([0, 1, 2, 3, 4])
        self.assertArrayEqual(series, cube.data)
        self.assertArrayEqual(series.index, expected_index)

    def test_simple(self):
        cube = Cube(np.array([0, 1, 2, 3, 4.4]), long_name="foo")
        dim_coord = DimCoord([5, 6, 7, 8, 9], long_name="bar")
        cube.add_dim_coord(dim_coord, 0)
        expected_index = np.array([5, 6, 7, 8, 9])
        series = iris.pandas.as_series(cube)
        self.assertArrayEqual(series, cube.data)
        self.assertArrayEqual(series.index, expected_index)

    def test_masked(self):
        data = np.ma.MaskedArray([0, 1, 2, 3, 4.4], mask=[0, 1, 0, 1, 0])
        cube = Cube(data, long_name="foo")
        series = iris.pandas.as_series(cube)
        self.assertArrayEqual(series, cube.data.astype("f").filled(np.nan))

    def test_time_gregorian(self):
        cube = Cube(np.array([0, 1, 2, 3, 4]), long_name="ts")
        time_coord = DimCoord(
            [0, 100.1, 200.2, 300.3, 400.4],
            long_name="time",
            units="days since 2000-01-01 00:00",
        )
        cube.add_dim_coord(time_coord, 0)
        expected_index = [
            datetime.datetime(2000, 1, 1, 0, 0),
            datetime.datetime(2000, 4, 10, 2, 24),
            datetime.datetime(2000, 7, 19, 4, 48),
            datetime.datetime(2000, 10, 27, 7, 12),
            datetime.datetime(2001, 2, 4, 9, 36),
        ]
        series = iris.pandas.as_series(cube)
        self.assertArrayEqual(series, cube.data)
        self.assertListEqual(list(series.index), expected_index)

    def test_time_360(self):
        cube = Cube(np.array([0, 1, 2, 3, 4]), long_name="ts")
        time_unit = cf_units.Unit(
            "days since 2000-01-01 00:00", calendar=cf_units.CALENDAR_360_DAY
        )
        time_coord = DimCoord(
            [0, 100.1, 200.2, 300.3, 400.4], long_name="time", units=time_unit
        )
        cube.add_dim_coord(time_coord, 0)
        expected_index = [
            cftime.Datetime360Day(2000, 1, 1, 0, 0),
            cftime.Datetime360Day(2000, 4, 11, 2, 24),
            cftime.Datetime360Day(2000, 7, 21, 4, 48),
            cftime.Datetime360Day(2000, 11, 1, 7, 12),
            cftime.Datetime360Day(2001, 2, 11, 9, 36),
        ]

        series = iris.pandas.as_series(cube)
        self.assertArrayEqual(series, cube.data)
        self.assertArrayEqual(series.index, expected_index)

    def test_copy_true(self):
        cube = Cube(np.array([0, 1, 2, 3, 4]), long_name="foo")
        series = iris.pandas.as_series(cube)
        series[0] = 99
        self.assertEqual(cube.data[0], 0)

    def test_copy_int32_false(self):
        cube = Cube(np.array([0, 1, 2, 3, 4], dtype=np.int32), long_name="foo")
        series = iris.pandas.as_series(cube, copy=False)
        series[0] = 99
        self.assertEqual(cube.data[0], 99)

    def test_copy_int64_false(self):
        cube = Cube(np.array([0, 1, 2, 3, 4], dtype=np.int32), long_name="foo")
        series = iris.pandas.as_series(cube, copy=False)
        series[0] = 99
        self.assertEqual(cube.data[0], 99)

    def test_copy_float_false(self):
        cube = Cube(np.array([0, 1, 2, 3.3, 4]), long_name="foo")
        series = iris.pandas.as_series(cube, copy=False)
        series[0] = 99
        self.assertEqual(cube.data[0], 99)

    def test_copy_masked_true(self):
        data = np.ma.MaskedArray([0, 1, 2, 3, 4], mask=[0, 1, 0, 1, 0])
        cube = Cube(data, long_name="foo")
        series = iris.pandas.as_series(cube)
        series[0] = 99
        self.assertEqual(cube.data[0], 0)

    def test_copy_masked_false(self):
        data = np.ma.MaskedArray([0, 1, 2, 3, 4], mask=[0, 1, 0, 1, 0])
        cube = Cube(data, long_name="foo")
        with self.assertRaises(ValueError):
            _ = iris.pandas.as_series(cube, copy=False)


@skip_pandas
class TestAsDataFrame(tests.IrisTest):
    """Test conversion of 2D cubes to Pandas using as_data_frame()"""

    def test_no_dim_coords(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        expected_index = [0, 1]
        expected_columns = [0, 1, 2, 3, 4]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_no_x_coord(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        y_coord = DimCoord([10, 11], long_name="bar")
        cube.add_dim_coord(y_coord, 0)
        expected_index = [10, 11]
        expected_columns = [0, 1, 2, 3, 4]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_no_y_coord(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        x_coord = DimCoord([10, 11, 12, 13, 14], long_name="bar")
        cube.add_dim_coord(x_coord, 1)
        expected_index = [0, 1]
        expected_columns = [10, 11, 12, 13, 14]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_simple(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        x_coord = DimCoord([10, 11, 12, 13, 14], long_name="bar")
        y_coord = DimCoord([15, 16], long_name="milk")
        cube.add_dim_coord(x_coord, 1)
        cube.add_dim_coord(y_coord, 0)
        expected_index = [15, 16]
        expected_columns = [10, 11, 12, 13, 14]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_masked(self):
        data = np.ma.MaskedArray(
            [[0, 1, 2, 3, 4.4], [5, 6, 7, 8, 9]],
            mask=[[0, 1, 0, 1, 0], [1, 0, 1, 0, 1]],
        )
        cube = Cube(data, long_name="foo")
        expected_index = [0, 1]
        expected_columns = [0, 1, 2, 3, 4]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data.astype("f").filled(np.nan))
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_time_gregorian(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="ts"
        )
        day_offsets = [0, 100.1, 200.2, 300.3, 400.4]
        time_coord = DimCoord(
            day_offsets, long_name="time", units="days since 2000-01-01 00:00"
        )
        cube.add_dim_coord(time_coord, 1)
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        nanoseconds_per_day = 24 * 60 * 60 * 1000000000
        days_to_2000 = 365 * 30 + 7
        # pandas Timestamp class cannot handle floats in pandas <v0.12
        timestamps = [
            pandas.Timestamp(
                int(nanoseconds_per_day * (days_to_2000 + day_offset))
            )
            for day_offset in day_offsets
        ]
        self.assertTrue(all(data_frame.columns == timestamps))
        self.assertTrue(all(data_frame.index == [0, 1]))

    def test_time_360(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="ts"
        )
        time_unit = cf_units.Unit(
            "days since 2000-01-01 00:00", calendar=cf_units.CALENDAR_360_DAY
        )
        time_coord = DimCoord(
            [100.1, 200.2], long_name="time", units=time_unit
        )
        cube.add_dim_coord(time_coord, 0)
        expected_index = [
            cftime.Datetime360Day(2000, 4, 11, 2, 24),
            cftime.Datetime360Day(2000, 7, 21, 4, 48),
        ]

        expected_columns = [0, 1, 2, 3, 4]
        data_frame = iris.pandas.as_data_frame(cube)
        self.assertArrayEqual(data_frame, cube.data)
        self.assertArrayEqual(data_frame.index, expected_index)
        self.assertArrayEqual(data_frame.columns, expected_columns)

    def test_copy_true(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        data_frame = iris.pandas.as_data_frame(cube)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 0)

    def test_copy_int32_false(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=np.int32),
            long_name="foo",
        )
        data_frame = iris.pandas.as_data_frame(cube, copy=False)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 99)

    def test_copy_int64_false(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]], dtype=np.int64),
            long_name="foo",
        )
        data_frame = iris.pandas.as_data_frame(cube, copy=False)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 99)

    def test_copy_float_false(self):
        cube = Cube(
            np.array([[0, 1, 2, 3, 4.4], [5, 6, 7, 8, 9]]), long_name="foo"
        )
        data_frame = iris.pandas.as_data_frame(cube, copy=False)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 99)

    def test_copy_masked_true(self):
        data = np.ma.MaskedArray(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            mask=[[0, 1, 0, 1, 0], [1, 0, 1, 0, 1]],
        )
        cube = Cube(data, long_name="foo")
        data_frame = iris.pandas.as_data_frame(cube)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 0)

    def test_copy_masked_false(self):
        data = np.ma.MaskedArray(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            mask=[[0, 1, 0, 1, 0], [1, 0, 1, 0, 1]],
        )
        cube = Cube(data, long_name="foo")
        with self.assertRaises(ValueError):
            _ = iris.pandas.as_data_frame(cube, copy=False)

    def test_copy_false_with_cube_view(self):
        data = np.array([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
        cube = Cube(data[:], long_name="foo")
        data_frame = iris.pandas.as_data_frame(cube, copy=False)
        data_frame[0][0] = 99
        self.assertEqual(cube.data[0, 0], 99)


@skip_pandas
class TestSeriesAsCube(tests.IrisTest):
    def test_series_simple(self):
        series = pandas.Series([0, 1, 2, 3, 4], index=[5, 6, 7, 8, 9])
        self.assertCML(
            iris.pandas.as_cube(series),
            tests.get_result_path(("pandas", "as_cube", "series_simple.cml")),
        )

    def test_series_object(self):
        class Thing:
            def __repr__(self):
                return "A Thing"

        series = pandas.Series(
            [0, 1, 2, 3, 4],
            index=[Thing(), Thing(), Thing(), Thing(), Thing()],
        )
        self.assertCML(
            iris.pandas.as_cube(series),
            tests.get_result_path(("pandas", "as_cube", "series_object.cml")),
        )

    def test_series_masked(self):
        series = pandas.Series(
            [0, float("nan"), 2, np.nan, 4], index=[5, 6, 7, 8, 9]
        )
        self.assertCML(
            iris.pandas.as_cube(series),
            tests.get_result_path(("pandas", "as_cube", "series_masked.cml")),
        )

    def test_series_datetime_gregorian(self):
        series = pandas.Series(
            [0, 1, 2, 3, 4],
            index=[
                datetime.datetime(2001, 1, 1, 1, 1, 1),
                datetime.datetime(2002, 2, 2, 2, 2, 2),
                datetime.datetime(2003, 3, 3, 3, 3, 3),
                datetime.datetime(2004, 4, 4, 4, 4, 4),
                datetime.datetime(2005, 5, 5, 5, 5, 5),
            ],
        )
        self.assertCML(
            iris.pandas.as_cube(series),
            tests.get_result_path(
                ("pandas", "as_cube", "series_datetime_gregorian.cml")
            ),
        )

    def test_series_cftime_360(self):
        series = pandas.Series(
            [0, 1, 2, 3, 4],
            index=[
                cftime.datetime(2001, 1, 1, 1, 1, 1),
                cftime.datetime(2002, 2, 2, 2, 2, 2),
                cftime.datetime(2003, 3, 3, 3, 3, 3),
                cftime.datetime(2004, 4, 4, 4, 4, 4),
                cftime.datetime(2005, 5, 5, 5, 5, 5),
            ],
        )
        self.assertCML(
            iris.pandas.as_cube(
                series, calendars={0: cf_units.CALENDAR_360_DAY}
            ),
            tests.get_result_path(
                ("pandas", "as_cube", "series_netcdfimte_360.cml")
            ),
        )

    def test_copy_true(self):
        series = pandas.Series([0, 1, 2, 3, 4], index=[5, 6, 7, 8, 9])
        cube = iris.pandas.as_cube(series)
        cube.data[0] = 99
        self.assertEqual(series[5], 0)

    def test_copy_false(self):
        series = pandas.Series([0, 1, 2, 3, 4], index=[5, 6, 7, 8, 9])
        cube = iris.pandas.as_cube(series, copy=False)
        cube.data[0] = 99
        self.assertEqual(series[5], 99)


@skip_pandas
class TestDataFrameAsCube(tests.IrisTest):
    def test_data_frame_simple(self):
        data_frame = pandas.DataFrame(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            index=[10, 11],
            columns=[12, 13, 14, 15, 16],
        )
        self.assertCML(
            iris.pandas.as_cube(data_frame),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_simple.cml")
            ),
        )

    def test_data_frame_nonotonic(self):
        data_frame = pandas.DataFrame(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            index=[10, 10],
            columns=[12, 12, 14, 15, 16],
        )
        self.assertCML(
            iris.pandas.as_cube(data_frame),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_nonotonic.cml")
            ),
        )

    def test_data_frame_masked(self):
        data_frame = pandas.DataFrame(
            [[0, float("nan"), 2, 3, 4], [5, 6, 7, np.nan, 9]],
            index=[10, 11],
            columns=[12, 13, 14, 15, 16],
        )
        self.assertCML(
            iris.pandas.as_cube(data_frame),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_masked.cml")
            ),
        )

    def test_data_frame_multidim(self):
        data_frame = pandas.DataFrame(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            index=[0, 1],
            columns=["col_1", "col_2", "col_3", "col_4", "col_5"],
        )
        self.assertCML(
            iris.pandas.as_cube(data_frame),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_multidim.cml")
            ),
        )

    def test_data_frame_cftime_360(self):
        data_frame = pandas.DataFrame(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            index=[
                cftime.datetime(2001, 1, 1, 1, 1, 1),
                cftime.datetime(2002, 2, 2, 2, 2, 2),
            ],
            columns=[10, 11, 12, 13, 14],
        )
        self.assertCML(
            iris.pandas.as_cube(
                data_frame, calendars={0: cf_units.CALENDAR_360_DAY}
            ),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_netcdftime_360.cml")
            ),
        )

    def test_data_frame_datetime_gregorian(self):
        data_frame = pandas.DataFrame(
            [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]],
            index=[
                datetime.datetime(2001, 1, 1, 1, 1, 1),
                datetime.datetime(2002, 2, 2, 2, 2, 2),
            ],
            columns=[10, 11, 12, 13, 14],
        )
        self.assertCML(
            iris.pandas.as_cube(data_frame),
            tests.get_result_path(
                ("pandas", "as_cube", "data_frame_datetime_gregorian.cml")
            ),
        )

    def test_copy_true(self):
        data_frame = pandas.DataFrame([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
        cube = iris.pandas.as_cube(data_frame)
        cube.data[0, 0] = 99
        self.assertEqual(data_frame[0][0], 0)

    def test_copy_false(self):
        data_frame = pandas.DataFrame([[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]])
        cube = iris.pandas.as_cube(data_frame, copy=False)
        cube.data[0, 0] = 99
        self.assertEqual(data_frame[0][0], 99)


if __name__ == "__main__":
    tests.main()



"""
Fitting a Polynomial
====================

This example demonstrates computing a polynomial fit to 1D data from an Iris
cube, adding the fit to the cube's metadata, and plotting both the 1D data and
the fit.

"""

import matplotlib.pyplot as plt
import numpy as np

import iris
import iris.quickplot as qplt


def main():
    # Load some test data.
    fname = iris.sample_data_path("A1B_north_america.nc")
    cube = iris.load_cube(fname)

    # Extract a single time series at a latitude and longitude point.
    location = next(cube.slices(["time"]))

    # Calculate a polynomial fit to the data at this time series.
    x_points = location.coord("time").points
    y_points = location.data
    degree = 2

    p = np.polyfit(x_points, y_points, degree)
    y_fitted = np.polyval(p, x_points)

    # Add the polynomial fit values to the time series to take
    # full advantage of Iris plotting functionality.
    long_name = "degree_{}_polynomial_fit_of_{}".format(degree, cube.name())
    fit = iris.coords.AuxCoord(
        y_fitted, long_name=long_name, units=location.units
    )
    location.add_aux_coord(fit, 0)

    qplt.plot(location.coord("time"), location, label="data")
    qplt.plot(
        location.coord("time"),
        location.coord(long_name),
        "g-",
        label="polynomial fit",
    )
    plt.legend(loc="best")
    plt.title("Trend of US air temperature over time")

    qplt.show()


if __name__ == "__main__":
    main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""Unit tests for the `iris.cube.CubeRepresentation` class."""

from html import escape

# Import iris.tests first so that some things can be initialised before
# importing anything else.
import iris.tests as tests

from iris.cube import CubeList
import iris.tests.stock as stock

from iris.experimental.representation import CubeListRepresentation


@tests.skip_data
class Test__instantiation(tests.IrisTest):
    def setUp(self):
        self.cubes = CubeList([stock.simple_3d()])
        self.representer = CubeListRepresentation(self.cubes)

    def test_ids(self):
        self.assertEqual(id(self.cubes), self.representer.cubelist_id)


@tests.skip_data
class Test_make_content(tests.IrisTest):
    def setUp(self):
        self.cubes = CubeList([stock.simple_3d(), stock.lat_lon_cube()])
        self.cubes[0].rename("name & <html>")
        self.representer = CubeListRepresentation(self.cubes)
        self.content = self.representer.make_content()

    def test_repr_len(self):
        self.assertEqual(len(self.cubes), len(self.content))

    def test_summary_lines(self):
        names = [c.name() for c in self.cubes]
        for name, content in zip(names, self.content):
            name = escape(name)
            self.assertIn(name, content)

    def test__cube_name_summary_consistency(self):
        # Just check the first cube in the CubeList.
        single_cube_html = self.content[0]
        # Get a "prettified" cube name, as it should be in the cubelist repr.
        cube_name = self.cubes[0].name()
        pretty_cube_name = cube_name.strip().replace("_", " ").title()
        pretty_escaped_name = escape(pretty_cube_name)
        self.assertIn(pretty_escaped_name, single_cube_html)


@tests.skip_data
class Test_repr_html(tests.IrisTest):
    def setUp(self):
        self.cubes = CubeList([stock.simple_3d(), stock.lat_lon_cube()])
        self.representer = CubeListRepresentation(self.cubes)

    def test_html_length(self):
        html = self.representer.repr_html()
        n_html_elems = html.count("<button")  # One <button> tag per cube.
        self.assertEqual(len(self.cubes), n_html_elems)


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

# Import Iris tests first so that some things can be initialised before
# importing anything else.

import iris.tests as tests

from . import gallerytest_util

with gallerytest_util.add_gallery_to_path():
    import plot_coriolis


class TestCoriolisPlot(tests.GraphicsTest):
    """Test the Coriolis Plot gallery code."""

    def test_plot_coriolis(self):
        with gallerytest_util.show_replaced_by_check_graphic(self):
            plot_coriolis.main()


if __name__ == "__main__":
    tests.main()



# Copyright Iris contributors
#
# This file is part of Iris and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

"""
Ugrid functions.

"""

import iris


def ugrid(location, name):
    """
    Create a cube from an unstructured grid.

    Args:

    * location:
        A string whose value represents the path to a file or
        URL to an OpenDAP resource conforming to the
        Unstructured Grid Metadata Conventions for Scientific Datasets
        https://github.com/ugrid-conventions/ugrid-conventions

    * name:
        A string whose value represents a cube loading constraint of
        first the standard name if found, then the long name if found,
        then the variable name if found, before falling back to
        the value of the default which itself defaults to "unknown"

    Returns:
        An instance of :class:`iris.cube.Cube` decorated with
        an instance of :class:`pyugrid.ugrid.Ugrid`
        bound to an attribute of the cube called "mesh"

    """
    # Lazy import so we can build the docs with no pyugrid.
    import pyugrid

    cube = iris.load_cube(location, name)
    ug = pyugrid.ugrid.UGrid.from_ncfile(location)
    cube.mesh = ug
    cube.mesh_dimension = 1  # {0:time, 1:node}
    return cube
